from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Final, Literal, Type, Union

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import azure, pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.util import AzureUrl
from great_expectations.datasource.fluent import _PandasFilePathDatasource
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
    _check_config_substitutions_needed,
)
from great_expectations.datasource.fluent.data_connector import (
    AzureBlobStorageDataConnector,
)
from great_expectations.datasource.fluent.interfaces import TestConnectionError
from great_expectations.datasource.fluent.pandas_datasource import PandasDatasourceError

if TYPE_CHECKING:
    from great_expectations.compatibility.azure import BlobServiceClient
    from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset

logger = logging.getLogger(__name__)


_MISSING: Final = object()


class PandasAzureBlobStorageDatasourceError(PandasDatasourceError):
    pass


@public_api
class PandasAzureBlobStorageDatasource(_PandasFilePathDatasource):
    """
    PandasAzureBlobStorageDatasource is a PandasDatasource that uses Azure Blob Storage as a
    data store.
    """

    # class attributes
    data_connector_type: ClassVar[Type[AzureBlobStorageDataConnector]] = (
        AzureBlobStorageDataConnector
    )

    # instance attributes
    type: Literal["pandas_abs"] = "pandas_abs"

    # Azure Blob Storage specific attributes
    azure_options: Dict[str, Union[ConfigStr, Any]] = {}

    _account_name: str = pydantic.PrivateAttr(default="")
    # on 3.11 the annotation must be type-checking import otherwise it will fail at import time
    _azure_client: Union[BlobServiceClient, None] = pydantic.PrivateAttr(default=None)

    def _get_azure_client(self) -> azure.BlobServiceClient:
        azure_client: Union[azure.BlobServiceClient, None] = self._azure_client
        if not azure_client:
            _check_config_substitutions_needed(
                self, self.azure_options, raise_warning_if_provider_not_present=True
            )
            # pull in needed config substitutions using the `_config_provider`
            # The `FluentBaseModel.dict()` call will do the config substitution on the serialized dict if a `config_provider` is passed.  # noqa: E501 # FIXME CoP
            azure_options: dict = self.dict(config_provider=self._config_provider).get(
                "azure_options", {}
            )

            # Thanks to schema validation, we are guaranteed to have one of `conn_str` or `account_url` to  # noqa: E501 # FIXME CoP
            # use in authentication (but not both). If the format or content of the provided keys is invalid,  # noqa: E501 # FIXME CoP
            # the assignment of `self._account_name` and `self._azure_client` will fail and an error will be raised.  # noqa: E501 # FIXME CoP
            conn_str: str | None = azure_options.get("conn_str")
            account_url: str | None = azure_options.get("account_url")
            if not bool(conn_str) ^ bool(account_url):
                raise PandasAzureBlobStorageDatasourceError(  # noqa: TRY003 # FIXME CoP
                    "You must provide one of `conn_str` or `account_url` to the `azure_options` key in your config (but not both)"  # noqa: E501 # FIXME CoP
                )

            # Validate that "azure" libararies were successfully imported and attempt to create "azure_client" handle.  # noqa: E501 # FIXME CoP
            if azure.BlobServiceClient:  # type: ignore[truthy-function] # False if NotImported
                try:
                    if conn_str is not None:
                        self._account_name = re.search(  # type: ignore[union-attr] # FIXME CoP
                            r".*?AccountName=(.+?);.*?", conn_str
                        ).group(1)
                        azure_client = azure.BlobServiceClient.from_connection_string(
                            **azure_options
                        )
                    elif account_url is not None:
                        self._account_name = re.search(  # type: ignore[union-attr] # FIXME CoP
                            r"(?:https?://)?(.+?).blob.core.windows.net", account_url
                        ).group(1)
                        azure_client = azure.BlobServiceClient(**azure_options)
                except Exception as e:
                    # Failure to create "azure_client" is most likely due invalid "azure_options" dictionary.  # noqa: E501 # FIXME CoP
                    raise PandasAzureBlobStorageDatasourceError(  # noqa: TRY003 # FIXME CoP
                        f'Due to exception: "{e!s}", "azure_client" could not be created.'
                    ) from e
            else:
                raise PandasAzureBlobStorageDatasourceError(  # noqa: TRY003 # FIXME CoP
                    'Unable to create "PandasAzureBlobStorageDatasource" due to missing azure.storage.blob dependency.'  # noqa: E501 # FIXME CoP
                )

            self._azure_client = azure_client

        if not azure_client:
            raise PandasAzureBlobStorageDatasourceError("Failed to return `azure_client`")  # noqa: TRY003 # FIXME CoP

        return azure_client

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the PandasAzureBlobStorageDatasource.

        Args:
            test_assets: If assets have been passed to the PandasAzureBlobStorageDatasource, whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """  # noqa: E501 # FIXME CoP
        try:
            _ = self._get_azure_client()
        except Exception as e:
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                f"Attempt to connect to datasource failed with the following error message: {e!s}"
            ) from e

        if self.assets and test_assets:
            for asset in self.assets:
                asset.test_connection()

    @override
    def _build_data_connector(
        self,
        data_asset: FileDataAsset,
        abs_container: str = _MISSING,  # type: ignore[assignment] # _MISSING is used as sentinel value
        abs_name_starts_with: str = "",
        abs_delimiter: str = "/",
        abs_recursive_file_discovery: bool = False,
        **kwargs,
    ) -> None:
        """Builds and attaches the `AzureBlobStorageDataConnector` to the asset."""
        if kwargs:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"_build_data_connector() got unexpected keyword arguments {list(kwargs.keys())}"
            )
        if abs_container is _MISSING:
            raise TypeError(f"'{data_asset.name}' is missing required argument 'abs_container'")  # noqa: TRY003 # FIXME CoP

        data_asset._data_connector = self.data_connector_type.build_data_connector(
            datasource_name=self.name,
            data_asset_name=data_asset.name,
            azure_client=self._get_azure_client(),
            account_name=self._account_name,
            container=abs_container,
            name_starts_with=abs_name_starts_with,
            delimiter=abs_delimiter,
            recursive_file_discovery=abs_recursive_file_discovery,
            file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
        )

        # build a more specific `_test_connection_error_message`
        data_asset._test_connection_error_message = (
            self.data_connector_type.build_test_connection_error_message(
                data_asset_name=data_asset.name,
                account_name=self._account_name,
                container=abs_container,
                name_starts_with=abs_name_starts_with,
                delimiter=abs_delimiter,
                recursive_file_discovery=abs_recursive_file_discovery,
            )
        )



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

EXISTING_COLUMN = "existing_column"
ANOTHER_EXISTING_COLUMN = "another_existing_column"

DATA = pd.DataFrame(
    {
        EXISTING_COLUMN: [1, 2],
        ANOTHER_EXISTING_COLUMN: ["a", "b"],
    }
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success_complete(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnToExist(
        column=EXISTING_COLUMN,
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnToExist(
                column=EXISTING_COLUMN,
            ),
            id="basic_column_exists",
        ),
        pytest.param(
            gxe.ExpectColumnToExist(
                column=ANOTHER_EXISTING_COLUMN,
            ),
            id="another_column_exists",
        ),
        pytest.param(
            gxe.ExpectColumnToExist(
                column=EXISTING_COLUMN,
                column_index=0,
            ),
            id="column_exists_at_index",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnToExist,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnToExist(
                column="non_existent_column",
            ),
            id="column_does_not_exist",
        ),
        pytest.param(
            gxe.ExpectColumnToExist(
                column=EXISTING_COLUMN,
                column_index=1,
            ),
            id="wrong_column_index",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnToExist,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success



from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING, Callable, Optional

from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    MissingFilePathTemplateMapFnError,
)
from great_expectations.datasource.fluent.data_connector.filesystem_data_connector import (
    FilesystemDataConnector,
)

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr

logger = logging.getLogger(__name__)


class DBFSDataConnector(FilesystemDataConnector):
    """Extension of FilePathDataConnector used to connect to the DataBricks File System (DBFS).

    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
        base_directory: Relative path to subdirectory containing files of interest
        glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
        data_context_root_directory: Optional GreatExpectations root directory (if installed on DBFS)
        file_path_template_map_fn: Format function mapping path to fully-qualified resource on DBFS
        get_unfiltered_batch_definition_list_fn: Function used to get the batch definition list before filtering
        whole_directory_path_override: If present, treat entire directory as single Asset
    """  # noqa: E501 # FIXME CoP

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_asset_name: str,
        base_directory: pathlib.Path,
        glob_directive: str = "**/*",
        data_context_root_directory: Optional[pathlib.Path] = None,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> None:
        super().__init__(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            base_directory=base_directory,
            glob_directive=glob_directive,
            data_context_root_directory=data_context_root_directory,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    @override
    def build_data_connector(  # noqa: PLR0913 # FIXME CoP
        cls,
        datasource_name: str,
        data_asset_name: str,
        base_directory: pathlib.Path,
        glob_directive: str = "**/*",
        data_context_root_directory: Optional[pathlib.Path] = None,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> DBFSDataConnector:
        """Builds "DBFSDataConnector", which links named DataAsset to DBFS.

        Args:
            datasource_name: The name of the Datasource associated with this "DBFSDataConnector" instance
            data_asset_name: The name of the DataAsset using this "DBFSDataConnector" instance
            base_directory: Relative path to subdirectory containing files of interest
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            data_context_root_directory: Optional GreatExpectations root directory (if installed on DBFS)
            file_path_template_map_fn: Format function mapping path to fully-qualified resource on DBFS
            get_unfiltered_batch_definition_list_fn: Function used to get the batch definition list before filtering
            whole_directory_path_override: If present, treat entire directory as single Asset

        Returns:
            Instantiated "DBFSDataConnector" object
        """  # noqa: E501 # FIXME CoP
        return DBFSDataConnector(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            base_directory=base_directory,
            glob_directive=glob_directive,
            data_context_root_directory=data_context_root_directory,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    # Interface Method
    @override
    def _get_full_file_path(self, path: str) -> str:
        if self._file_path_template_map_fn is None:
            raise MissingFilePathTemplateMapFnError()

        template_arguments = {
            "path": str(self.base_directory.joinpath(path)),
        }

        return self._file_path_template_map_fn(**template_arguments)



from datetime import datetime
from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
    NON_SQL_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

NUMBERS_A = "numbers_a"
NUMBERS_A_EQUAL = "numbers_a_equal"
NUMBERS_B = "numbers_b"
DATES_A = "dates_a"
DATES_B = "dates_b"
STRINGS_A = "strings_a"
STRINGS_B = "strings_b"
NULLS_A = "nulls_a"
NULLS_B = "nulls_b"
NULLS_C = "nulls_c"

DATA = pd.DataFrame(
    {
        NUMBERS_A: [5, 8, 11, 13, 14],
        NUMBERS_A_EQUAL: [5, 8, 11, 13, 14],
        NUMBERS_B: [1, 5, 9, 12, 12],
        DATES_A: [
            datetime(2024, 1, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 2, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 3, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 4, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 5, 1).date(),  # noqa: DTZ001 # FIXME CoP
        ],
        DATES_B: [
            datetime(2023, 12, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 1, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 2, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 3, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 4, 1).date(),  # noqa: DTZ001 # FIXME CoP
        ],
        STRINGS_A: ["b", "m", "y", "z", "zz"],
        STRINGS_B: ["a", "k", "x", "y", "za"],
        NULLS_A: [None, None, None, None, 1],
        NULLS_B: [1, 2, None, 4, 5],
        NULLS_C: [2, 3, None, 5, 6],
    },
    dtype="object",
)


@parameterize_batch_for_data_sources(data_source_configs=NON_SQL_DATA_SOURCES, data=DATA)
def test_success_complete_non_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnPairValuesAToBeGreaterThanB(
        column_A=NUMBERS_A,
        column_B=NUMBERS_B,
        or_equal=False,
        ignore_row_if="either_value_is_missing",
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=SQL_DATA_SOURCES, data=DATA)
def test_success_complete_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnPairValuesAToBeGreaterThanB(
        column_A=NUMBERS_A,
        column_B=NUMBERS_B,
        or_equal=False,
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 5,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
        "missing_count": 0,
        "missing_percent": 0,
        "unexpected_percent_total": 0.0,
        "unexpected_percent_nonmissing": 0.0,
        "partial_unexpected_counts": [],
        "unexpected_list": [],
        "unexpected_index_query": ANY,
    }


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NUMBERS_A,
                column_B=NUMBERS_B,
            ),
            id="basic_number_comparison",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NUMBERS_A,
                column_B=NUMBERS_A_EQUAL,
                or_equal=True,
            ),
            id="number_comparison_with_equality",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=DATES_A,
                column_B=DATES_B,
            ),
            id="date_comparison",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=STRINGS_A,
                column_B=STRINGS_B,
            ),
            id="string_comparison",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=STRINGS_A,
                column_B=STRINGS_B,
            ),
            id="string_comparison_with_equality",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NULLS_C,
                column_B=NULLS_B,
                ignore_row_if="both_values_are_missing",
            ),
            id="ignore_nulls_both",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NUMBERS_A,
                column_B=NUMBERS_B,
            ),
            id="mostly_success",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnPairValuesAToBeGreaterThanB,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NUMBERS_A,
                column_B=NUMBERS_A_EQUAL,
                or_equal=False,
            ),
            id="numbers_are_equal",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NUMBERS_B,
                column_B=NUMBERS_A,
            ),
            id="reversed_columns",
        ),
        pytest.param(
            gxe.ExpectColumnPairValuesAToBeGreaterThanB(
                column_A=NULLS_A,
                column_B=NULLS_B,
                ignore_row_if="neither",
            ),
            id="null_comparison_not_ignored",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnPairValuesAToBeGreaterThanB,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_or_equal_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_pair_values_a_to_be_greater_than_b"
    expectation = gxe.ExpectColumnPairValuesAToBeGreaterThanB(
        column_A=NUMBERS_A,
        column_B=NUMBERS_B,
        or_equal={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param("both_values_are_missing", True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_ignore_row_if_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_pair_values_a_to_be_greater_than_b"
    expectation = gxe.ExpectColumnPairValuesAToBeGreaterThanB(
        column_A=NUMBERS_A,
        column_B=NUMBERS_B,
        ignore_row_if={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from typing import TYPE_CHECKING

import pytest
from contrib.experimental.great_expectations_experimental.expectations.expect_queried_table_row_count_to_be import (  # noqa: E501 # FIXME CoP
    ExpectQueriedTableRowCountToBe,  # noqa: F401 # needed for expectation registration
)

# noinspection PyUnresolvedReferences
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest
from great_expectations.self_check.util import get_test_validator_with_data
from great_expectations.util import build_in_memory_runtime_context
from great_expectations.validator.validator import (
    ExpectationValidationResult,
    Validator,
)

if TYPE_CHECKING:
    import pandas as pd

sqlite_runtime_batch_request: RuntimeBatchRequest = RuntimeBatchRequest(
    datasource_name="my_sqlite_db_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="titanic",
    runtime_parameters={"query": "SELECT * FROM titanic LIMIT 100"},
    batch_identifiers={"default_identifier_name": "test_identifier"},
    batch_spec_passthrough={"create_temp_table": False},
)

sqlite_batch_request: BatchRequest = BatchRequest(
    datasource_name="my_sqlite_db_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="titanic",
    batch_spec_passthrough={"create_temp_table": False},
)

pytest.skip("TODO: Fix in V1-323", allow_module_level=True)


@pytest.mark.parametrize(
    "batch_request,success,value,observed,row_condition",
    [
        (sqlite_runtime_batch_request, True, 100, 100, None),
        (sqlite_batch_request, False, 100, 1313, None),
        (sqlite_batch_request, False, 100, 96, 'col("Age")<18'),
        (sqlite_runtime_batch_request, True, 70, 70, 'col("Age")>17'),
    ],
)
@pytest.mark.slow  # 4.32s
@pytest.mark.big
def test_expect_queried_column_value_frequency_to_meet_threshold_sqlite(
    batch_request,
    success,
    value,
    observed,
    row_condition,
    titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    validator: Validator = context.get_validator(batch_request=batch_request)

    result: ExpectationValidationResult = validator.expect_queried_table_row_count_to_be(
        value=value,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )

    assert result["success"] is success and result["result"]["observed_value"] == observed


@pytest.mark.parametrize(
    "batch_request,success,query,value,observed,row_condition",
    [
        (
            sqlite_runtime_batch_request,
            True,
            "SELECT COUNT(*) FROM titanic",
            1313,
            1313,
            None,
        ),
        (
            sqlite_batch_request,
            True,
            "SELECT COUNT (*) FROM (SELECT * FROM titanic LIMIT 100)",
            100,
            100,
            'col("Age")>17',
        ),
    ],
)
@pytest.mark.slow  # 1.59s
@pytest.mark.big
def test_expect_queried_column_value_frequency_to_meet_threshold_override_query_sqlite(
    batch_request,
    success,
    query,
    value,
    observed,
    row_condition,
    titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    validator: Validator = context.get_validator(batch_request=batch_request)

    result: ExpectationValidationResult = validator.expect_queried_table_row_count_to_be(
        value=value,
        query=query,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )

    assert result["success"] is success and result["result"]["observed_value"] == observed


# noinspection PyUnusedLocal
@pytest.mark.parametrize(
    "success,value,observed,row_condition",
    [
        (False, 100, 1313, None),
        (False, 100, 96, 'col("Age")<18'),
    ],
)
@pytest.mark.spark
def test_expect_queried_column_value_frequency_to_meet_threshold_spark(
    success,
    value,
    observed,
    row_condition,
    spark_session,
    basic_spark_df_execution_engine,
    titanic_df,
):
    df: pd.DataFrame = titanic_df

    context = build_in_memory_runtime_context()

    validator = get_test_validator_with_data(
        execution_engine="spark",
        data=df,
        context=context,
    )

    result: ExpectationValidationResult = validator.expect_queried_table_row_count_to_be(
        value=value,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )

    assert result["success"] is success and result["result"]["observed_value"] == observed


# noinspection PyUnusedLocal
@pytest.mark.parametrize(
    "success,query,value,observed,row_condition",
    [
        (
            True,
            "SELECT COUNT (*) FROM (SELECT * FROM {batch} LIMIT 100)",
            100,
            100,
            'col("Age")>17',
        ),
    ],
)
@pytest.mark.spark
def test_expect_queried_column_value_frequency_to_meet_threshold_override_query_spark(
    success,
    query,
    value,
    observed,
    row_condition,
    spark_session,
    titanic_df,
):
    df: pd.DataFrame = titanic_df

    context = build_in_memory_runtime_context()

    validator = get_test_validator_with_data(
        execution_engine="spark",
        data=df,
        context=context,
    )

    result: ExpectationValidationResult = validator.expect_queried_table_row_count_to_be(
        value=value,
        query=query,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )

    assert result["success"] is success and result["result"]["observed_value"] == observed



from enum import Enum


class DataQualityIssues(str, Enum):
    """Data quality issues addressed by Core Expectations."""

    VOLUME = "Volume"
    SCHEMA = "Schema"
    COMPLETENESS = "Completeness"
    UNIQUENESS = "Uniqueness"
    NUMERIC = "Numeric"
    VALIDITY = "Validity"
    SQL = "SQL"
    MULTI_SOURCE = "Multi-source"


class SupportedDataSources(str, Enum):
    """Data sources supported by Core Expectations."""

    PANDAS = "Pandas"
    SPARK = "Spark"
    SQLITE = "SQLite"
    POSTGRESQL = "PostgreSQL"
    MYSQL = "MySQL"
    MSSQL = "MSSQL"
    BIGQUERY = "BigQuery"
    SNOWFLAKE = "Snowflake"
    DATABRICKS = "Databricks (SQL)"
    REDSHIFT = "Redshift"



import pytest

from great_expectations import expectations as gxe
from great_expectations.render import (
    RenderedAtomicContent,
    RenderedAtomicValue,
)
from great_expectations.render.renderer.inline_renderer import InlineRenderer
from great_expectations.render.renderer_configuration import CodeBlock, CodeBlockLanguage


@pytest.mark.parametrize(
    "expectation,expected_expectation_configuration_rendered_atomic_content",
    [
        pytest.param(
            gxe.ExpectQueryResultsToMatchComparison(
                description="Both tables should be identical",
                base_query="SELECT * FROM {batch}",
                comparison_data_source_name="My Data Source",
                comparison_query="SELECT * FROM a_table_in_source_data_source",
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template="Both tables should be identical",
                        code_block=CodeBlock(
                            code_template_str="$base_query", language=CodeBlockLanguage.SQL
                        ),
                        params={
                            "base_query": {
                                "schema": {"type": "string"},
                                "value": "SELECT * FROM {batch}",
                            }
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template="Compare with Data Source $comparison_data_source_name",
                        code_block=CodeBlock(
                            code_template_str="$comparison_query", language=CodeBlockLanguage.SQL
                        ),
                        params={
                            "comparison_data_source_name": {
                                "schema": {"type": "string"},
                                "value": "My Data Source",
                            },
                            "comparison_query": {
                                "schema": {"type": "string"},
                                "value": "SELECT * FROM a_table_in_source_data_source",
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="with_description",
        ),
        pytest.param(
            gxe.ExpectQueryResultsToMatchComparison(
                base_query="SELECT * FROM {batch}",
                comparison_data_source_name="My Data Source",
                comparison_query="SELECT * FROM a_table_in_source_data_source",
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        code_block=CodeBlock(
                            code_template_str="$base_query", language=CodeBlockLanguage.SQL
                        ),
                        params={
                            "base_query": {
                                "schema": {"type": "string"},
                                "value": "SELECT * FROM {batch}",
                            }
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template="Compare with Data Source $comparison_data_source_name",
                        code_block=CodeBlock(
                            code_template_str="$comparison_query", language=CodeBlockLanguage.SQL
                        ),
                        params={
                            "comparison_data_source_name": {
                                "schema": {"type": "string"},
                                "value": "My Data Source",
                            },
                            "comparison_query": {
                                "schema": {"type": "string"},
                                "value": "SELECT * FROM a_table_in_source_data_source",
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="no_description",
        ),
    ],
)
@pytest.mark.unit
def test_expectation_configuration_rendered_atomic_content(
    expectation: gxe.Expectation,
    expected_expectation_configuration_rendered_atomic_content: dict,
):
    inline_renderer: InlineRenderer = InlineRenderer(render_object=expectation.configuration)

    actual_expectation_configuration_rendered_atomic_content: list[RenderedAtomicContent] = (
        inline_renderer.get_rendered_content()
    )

    assert len(actual_expectation_configuration_rendered_atomic_content) == 2

    assert (
        expected_expectation_configuration_rendered_atomic_content
        == actual_expectation_configuration_rendered_atomic_content
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, Union

import marshmallow
from ruamel.yaml import YAML

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.store.store import Store
from great_expectations.data_context.store.tuple_store_backend import TupleStoreBackend
from great_expectations.data_context.types.base import BaseYamlConfig
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
    GXCloudIdentifier,
)
from great_expectations.data_context.util import load_class
from great_expectations.util import (
    filter_properties_dict,
    verify_dynamic_loading_support,
)

if TYPE_CHECKING:
    from ruamel.yaml.comments import CommentedMap

yaml = YAML()

yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False

logger = logging.getLogger(__name__)


class ConfigurationStore(Store):
    """
    Configuration Store provides a way to store any Marshmallow Schema compatible Configuration (using the YAML format).
    """  # noqa: E501 # FIXME CoP

    _key_class = ConfigurationIdentifier

    _configuration_class = BaseYamlConfig

    def __init__(
        self,
        store_name: str,
        store_backend: Optional[dict] = None,
        overwrite_existing: bool = False,
        runtime_environment: Optional[dict] = None,
    ) -> None:
        if not issubclass(self._configuration_class, BaseYamlConfig):
            raise gx_exceptions.DataContextError(  # noqa: TRY003 # FIXME CoP
                "Invalid configuration: A configuration_class needs to inherit from the BaseYamlConfig class."  # noqa: E501 # FIXME CoP
            )

        if store_backend is not None:
            store_backend_module_name = store_backend.get(
                "module_name", "great_expectations.data_context.store"
            )
            store_backend_class_name = store_backend.get("class_name", "InMemoryStoreBackend")
            verify_dynamic_loading_support(module_name=store_backend_module_name)
            store_backend_class = load_class(store_backend_class_name, store_backend_module_name)

            # Store Backend Class was loaded successfully; verify that it is of a correct subclass.
            if issubclass(store_backend_class, TupleStoreBackend):
                # Provide defaults for this common case
                store_backend["filepath_suffix"] = store_backend.get("filepath_suffix", ".yml")

        super().__init__(
            store_backend=store_backend,
            runtime_environment=runtime_environment,
            store_name=store_name,
        )

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "store_name": store_name,
            "store_backend": store_backend,
            "overwrite_existing": overwrite_existing,
            "runtime_environment": runtime_environment,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

        self._overwrite_existing = overwrite_existing

    def serialize(self, value):  # type: ignore[explicit-override] # FIXME
        if self.cloud_mode:
            # GXCloudStoreBackend expects a json str
            config_schema = value.get_schema_class()()
            return config_schema.dump(value)
        return value.to_yaml_str()

    def deserialize(self, value):  # type: ignore[explicit-override] # FIXME
        config = value
        if isinstance(value, str):
            config: CommentedMap = yaml.load(value)
        try:
            return self._configuration_class.from_commented_map(commented_map=config)
        except gx_exceptions.InvalidBaseYamlConfigError:
            # Just to be explicit about what we intended to catch
            raise
        except marshmallow.ValidationError as e:
            raise gx_exceptions.InvalidBaseYamlConfigError(  # noqa: TRY003 # FIXME CoP
                f"Deserialized configuration failed validation: {e}"
            )

    @property
    def overwrite_existing(self) -> bool:
        return self._overwrite_existing

    @overwrite_existing.setter
    def overwrite_existing(self, overwrite_existing: bool) -> None:
        self._overwrite_existing = overwrite_existing

    @property
    @override
    def config(self) -> dict:
        return self._config

    def get_key(
        self, name: Optional[str] = None, id: Optional[str] = None
    ) -> Union[GXCloudIdentifier, ConfigurationIdentifier]:
        assert bool(name) ^ bool(id), "Must provide either name or id."

        key: Union[GXCloudIdentifier, ConfigurationIdentifier]
        if id or self.cloud_mode:
            key = GXCloudIdentifier(
                resource_type=GXCloudRESTResource.CHECKPOINT,
                id=id,
                resource_name=name,
            )
        else:
            key = ConfigurationIdentifier(configuration_key=name)  # type: ignore[arg-type] # FIXME CoP

        return key



import pandas as pd

from tests.integration.db.taxi_data_utils import (
    _execute_taxi_partitioning_test_cases,
    loaded_table,
)
from tests.integration.fixtures.partition_and_sample_data.partitioner_test_cases_and_fixtures import (  # noqa: E501 # FIXME CoP
    TaxiPartitioningTestCasesBase,
    TaxiPartitioningTestCasesDateTime,
    TaxiTestData,
)
from tests.test_utils import get_connection_string_and_dialect

if __name__ == "test_script_module":
    dialect: str
    connection_string: str
    dialect, connection_string = get_connection_string_and_dialect(
        athena_db_name_env_var="ATHENA_TEN_TRIPS_DB_NAME"
    )
    print(f"Testing dialect: {dialect}")

    with loaded_table(dialect=dialect, connection_string=connection_string) as table:
        table_name: str = table.table_name
        test_df: pd.DataFrame = table.inserted_dataframe

        test_column_name: str = "pickup_datetime"

        taxi_test_data: TaxiTestData = TaxiTestData(
            test_df=test_df,
            test_column_name=test_column_name,
            test_column_names=None,
        )
        taxi_partitioning_test_cases: TaxiPartitioningTestCasesBase = (
            TaxiPartitioningTestCasesDateTime(taxi_test_data=taxi_test_data)
        )
        _execute_taxi_partitioning_test_cases(
            taxi_partitioning_test_cases=taxi_partitioning_test_cases,
            connection_string=connection_string,
            table_name=table_name,
        )



from great_expectations.compatibility.pydantic import BaseModel
from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class DescriptiveStats(BaseModel):
    min: float
    max: float
    mean: float
    standard_deviation: float


class ColumnDescriptiveStatsResult(MetricResult[DescriptiveStats]): ...


class ColumnDescriptiveStats(ColumnMetric[ColumnDescriptiveStatsResult]):
    """Summary statistics for a column: min, mean, max, standard deviation"""

    name = "column.descriptive_stats"



from great_expectations.compatibility.not_imported import NotImported

PY4J_NOT_IMPORTED = NotImported("py4j is not installed, please 'pip install py4j'")

try:
    from py4j import protocol  # type: ignore[import-untyped] # FIXME CoP
except ImportError:
    protocol = PY4J_NOT_IMPORTED



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Set,
)

if TYPE_CHECKING:
    from great_expectations.expectations.metrics import MetaMetricProvider

logger = logging.getLogger(__name__)

SQLALCHEMY_SELECTABLE_METRICS: Set[str] = {
    "compound_columns.count",
    "compound_columns.unique",
}


def _is_sqlalchemy_metric_selectable(
    map_metric_provider: MetaMetricProvider,
) -> bool:
    """
    :param map_metric_provider: object of type "MapMetricProvider", whose SQLAlchemy implementation is inspected
    :return: boolean indicating whether or not the returned value of a method implementing the metric resolves all
    columns -- hence the caller must not use "select_from" clause as part of its own SQLAlchemy query; otherwise an
    unwanted selectable (e.g., table) will be added to "FROM", leading to duplicated and/or erroneous results.
    """  # noqa: E501 # FIXME CoP
    # noinspection PyUnresolvedReferences
    return (
        hasattr(map_metric_provider, "condition_metric_name")
        and map_metric_provider.condition_metric_name in SQLALCHEMY_SELECTABLE_METRICS
    ) or (
        hasattr(map_metric_provider, "function_metric_name")
        and map_metric_provider.function_metric_name in SQLALCHEMY_SELECTABLE_METRICS
    )



from __future__ import annotations

import pytest

from great_expectations.compatibility import pydantic
from great_expectations.datasource.fluent.databricks_sql_datasource import (
    DatabricksSQLDatasource,
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "connection_string, expected_errors",
    [
        pytest.param(
            "databricks://token:my_token>@my_host:1234",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",  # noqa: E501 # FIXME CoP
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "URL query is invalid or missing",
                    "type": "value_error.url.query",
                },
            ],
            id="missing query",
        ),
        pytest.param(
            "databricks://token:my_token>@my_host:1234?my_query=data",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",  # noqa: E501 # FIXME CoP
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "'http_path' query param is invalid or missing",
                    "type": "value_error.url.query.http_path",
                },
            ],
            id="missing http_path",
        ),
        pytest.param(
            "databricks://token:my_token>@my_host:1234?http_path=/path/a/&http_path=/path/b/",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",  # noqa: E501 # FIXME CoP
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "Only one `http_path` query entry is allowed",
                    "type": "value_error",
                },
            ],
            id="multiple http_paths",
        ),
        pytest.param(
            "databricks://token:my_token>@my_host:1234?http_path=/a/b/c&schema=dev",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",  # noqa: E501 # FIXME CoP
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "'catalog' query param is invalid or missing",
                    "type": "value_error.url.query.catalog",
                },
            ],
            id="missing catalog",
        ),
        pytest.param(
            "databricks://token:my_token>@my_host:1234?http_path=/a/b/c&catalog=dev",
            [
                {
                    "loc": ("connection_string",),
                    "msg": "ConfigStr - contains no config template strings in the format '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'",  # noqa: E501 # FIXME CoP
                    "type": "value_error",
                },
                {
                    "loc": ("connection_string",),
                    "msg": "'schema' query param is invalid or missing",
                    "type": "value_error.url.query.schema",
                },
            ],
            id="missing schema",
        ),
    ],
)
def test_invalid_connection_string_raises_dsn_error(
    connection_string: str, expected_errors: list[dict]
):
    with pytest.raises(pydantic.ValidationError) as exc_info:
        _ = DatabricksSQLDatasource(name="my_databricks", connection_string=connection_string)

    assert expected_errors == exc_info.value.errors()
    assert "my_token" not in str(exc_info.value.errors())



from __future__ import annotations

import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Final,
    List,
    NoReturn,
    Type,
    Union,
    overload,
)

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent import (
    DataAsset,
    Datasource,
    GxDatasourceWarning,
    TestConnectionError,
)
from great_expectations.datasource.fluent.type_lookup import TypeLookup, ValidTypes

if TYPE_CHECKING:
    from great_expectations.core.partitioners import ColumnPartitioner
    from great_expectations.datasource.fluent.batch_request import BatchRequest
    from great_expectations.datasource.fluent.interfaces import (
        Batch,
        PartitionerSortingProtocol,
    )

# Controls which methods should raise an error when called on an InvalidDatasource
METHOD_SHOULD_RAISE_ERROR: Final[set] = {
    "get_batch",
    "get_batch_identifiers_list",
    "add_batch_definition",
}


class GxInvalidDatasourceWarning(GxDatasourceWarning):
    """
    A warning that the Datasource configuration is invalid and will must be updated before it can used.
    """  # noqa: E501 # FIXME CoP


class InvalidAsset(DataAsset):
    """
    A DataAsset that is invalid.
    The DataAsset itself may be valid, but it is classified as invalid because its parent Datasource or sibling assets are invalid.
    """  # noqa: E501 # FIXME CoP

    type: str = "invalid"
    name: str = "invalid"

    class Config:
        extra = "ignore"

    def _raise_type_error(self) -> NoReturn:
        """
        Raise a TypeError indicating that the Asset is invalid.
        If available, raise from the original config error that caused the Datasource to be invalid.
        """
        error = TypeError(f"{self.name} Asset is invalid")
        if datasource := getattr(self, "datasource", None):
            raise error from datasource.config_error
        raise error

    @override
    def test_connection(self) -> None:
        if datasource := getattr(self, "datasource", None):
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                f"The Datasource configuration for {self.name} is invalid and cannot be used. Please fix the error and try again"  # noqa: E501 # FIXME CoP
            ) from datasource.config_error
        # the asset should always have a datasource, but if it doesn't, we should still raise an error  # noqa: E501 # FIXME CoP
        raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
            "This Asset configuration is invalid and cannot be used. Please fix the error and try again"  # noqa: E501 # FIXME CoP
        )

    @override
    def add_batch_definition(self, name: str, partitioner: Any | None = None) -> NoReturn:
        self._raise_type_error()

    @override
    def build_batch_request(
        self,
        options: dict | None = None,
        batch_slice: Any = None,
        partitioner: Any = None,
    ) -> NoReturn:
        self._raise_type_error()

    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        self._raise_type_error()

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        self._raise_type_error()

    @override
    def sort_batches(
        self, batch_list: List[Batch], partitioner: PartitionerSortingProtocol
    ) -> List[Batch]:
        self._raise_type_error()

    @override
    def get_batch_parameters_keys(self, partitioner: ColumnPartitioner | None = None) -> NoReturn:
        self._raise_type_error()


class InvalidAssetTypeLookup(TypeLookup):
    """A TypeLookup that always returns InvalidAsset for any type."""

    @overload
    def __getitem__(self, key: str) -> Type: ...

    @overload
    def __getitem__(self, key: Type) -> str: ...

    @override
    def __getitem__(self, key: ValidTypes) -> ValidTypes:
        if isinstance(key, str):
            return InvalidAsset
        # if a type is passed, normally we would return the type name but that doesn't make sense here  # noqa: E501 # FIXME CoP
        # for an InvalidAsset
        raise NotImplementedError(
            f"Looking up the `type` name for {InvalidAsset.__name__} is not supported"
        )


class InvalidDatasource(Datasource):
    """
    A Datasource that is invalid.

    This is used to represent a Datasource that is invalid and cannot be used.

    This class should override all methods that would commonly be called when a user intends to use the Datasource.
    The overridden methods should indicate to the user that the Datasource configuration is invalid and provide details about
    why it was considered invalid.

    Any errors raised should raise `from self.config_error`.
    """  # noqa: E501 # FIXME CoP

    # class var definitions
    asset_types: ClassVar[List[Type[DataAsset]]] = [InvalidAsset]
    _type_lookup: ClassVar[TypeLookup] = InvalidAssetTypeLookup()

    type: str = "invalid"
    config_error: Union[pydantic.ValidationError, LookupError] = Field(
        ..., description="The error that caused the Datasource to be invalid."
    )
    assets: List[InvalidAsset] = []

    class Config:
        extra = "ignore"
        arbitrary_types_allowed = True
        json_encoders = {
            pydantic.ValidationError: lambda v: v.errors(),
            LookupError: lambda v: repr(v),
        }

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
            "This Datasource configuration is invalid and cannot be used. Please fix the error and try again"  # noqa: E501 # FIXME CoP
        ) from self.config_error

    @override
    def get_asset(self, name: str) -> InvalidAsset:
        """
        Always raise a warning and return an InvalidAsset.
        Don't raise an error because the users may want to inspect the asset config.
        """
        warnings.warn(
            f"The {self.name} Datasource configuration is invalid and cannot be used. Please fix the error and try again",  # noqa: E501 # FIXME CoP
            GxInvalidDatasourceWarning,
        )
        return super().get_asset(name)

    def _raise_type_error(self, *args, **kwargs) -> NoReturn:
        """
        Raise a TypeError indicating that the Datasource is invalid.
        Raise from the original config error that caused the Datasource to be invalid.
        """
        error = TypeError(
            f"{self.name} Datasource is configuration is invalid and cannot be used. Please fix the error and try again"  # noqa: E501 # FIXME CoP
        )
        raise error from self.config_error

    @override
    def __getattribute__(self, attr: str):
        """
        Dynamically raise a TypeError with details of the original config error for
        any methods and attributes that do not make sense for an InvalidDatasource.
        """
        if attr in METHOD_SHOULD_RAISE_ERROR:
            raise AttributeError  # this causes __getattr__ to be called
        return super().__getattribute__(attr)

    def __getattr__(self, attr: str):
        # __getattr__ is only called if the attribute is not found by __getattribute__
        if attr in ("add_dataframe_asset", "__deepcopy__"):
            # these methods are part of protocol checks and should return None
            return None
        return self._raise_type_error()



from __future__ import annotations

import json
import logging
import uuid
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from marshmallow import Schema, fields, post_dump, post_load, pre_dump

import great_expectations.exceptions as gx_exceptions
from great_expectations import __version__ as ge_version
from great_expectations._docs_decorators import (
    public_api,
)
from great_expectations.analytics.anonymizer import anonymize
from great_expectations.analytics.client import submit as submit_event
from great_expectations.analytics.events import (
    ExpectationSuiteExpectationCreatedEvent,
    ExpectationSuiteExpectationDeletedEvent,
    ExpectationSuiteExpectationUpdatedEvent,
)
from great_expectations.compatibility.pydantic import ValidationError as PydanticValidationError
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.freshness_diagnostics import (
    ExpectationSuiteFreshnessDiagnostics,
)
from great_expectations.core.serdes import _IdentifierBundle
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.exceptions import (
    ExpectationSuiteError,
    ExpectationSuiteNotAddedError,
    ExpectationSuiteNotFoundError,
    ExpectationSuiteNotFreshError,
    StoreBackendError,
)
from great_expectations.exceptions.exceptions import InvalidKeyError
from great_expectations.types import SerializableDictDot
from great_expectations.util import (
    convert_to_json_serializable,  # noqa: TID251 # FIXME CoP
    ensure_json_serializable,  # noqa: TID251 # FIXME CoP
)

if TYPE_CHECKING:
    from great_expectations.alias_types import JSONValues
    from great_expectations.data_context.store.expectations_store import ExpectationsStore
    from great_expectations.expectations.expectation import Expectation
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )

    _TExpectation = TypeVar("_TExpectation", bound=Expectation)

logger = logging.getLogger(__name__)


@public_api
class ExpectationSuite(SerializableDictDot):
    """Set-like collection of Expectations.

    Args:
        name: Name of the Expectation Suite
        expectations: Expectation Configurations to associate with this Expectation Suite.
        suite_parameters: Suite parameters to be substituted when evaluating Expectations.
        meta: Metadata related to the suite.
        id: Great Expectations Cloud id for this Expectation Suite.
    """

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        name: Optional[str] = None,
        expectations: Optional[Sequence[Union[dict, ExpectationConfiguration, Expectation]]] = None,
        suite_parameters: Optional[dict] = None,
        meta: Optional[dict] = None,
        notes: str | list[str] | None = None,
        id: Optional[str] = None,
    ) -> None:
        if not name or not isinstance(name, str):
            raise ValueError("name must be provided as a non-empty string")  # noqa: TRY003 # FIXME CoP
        self.name = name
        self.id = id

        self.expectations = []
        for exp in expectations or []:
            try:
                self.expectations.append(self._process_expectation(exp))
            except gx_exceptions.InvalidExpectationConfigurationError as e:
                logger.exception(
                    f"Could not add expectation; provided configuration is not valid: {e.message}"
                )

        if suite_parameters is None:
            suite_parameters = {}
        self.suite_parameters = suite_parameters
        if meta is None:
            meta = {"great_expectations_version": ge_version}
        if (
            "great_expectations.__version__" not in meta
            and "great_expectations_version" not in meta
        ):
            meta["great_expectations_version"] = ge_version
        # We require meta information to be serializable, but do not convert until necessary
        ensure_json_serializable(meta)
        self.meta = meta
        self.notes = notes

    @property
    def _store(self) -> ExpectationsStore:
        return project_manager.get_expectations_store()

    @property
    def _include_rendered_content(self) -> bool:
        return project_manager.is_using_cloud()

    @property
    def suite_parameter_options(self) -> tuple[str, ...]:
        """SuiteParameter options for this ExpectationSuite.

        Returns:
            tuple[str, ...]: The keys of the suite parameters used by all Expectations of this suite at runtime.
        """  # noqa: E501 # FIXME CoP
        output: set[str] = set()
        for expectation in self.expectations:
            output.update(expectation.suite_parameter_options)
        return tuple(sorted(output))

    @public_api
    def add_expectation(self, expectation: _TExpectation) -> _TExpectation:
        """Add an Expectation to the collection."""
        if expectation.id:
            raise RuntimeError(  # noqa: TRY003 # FIXME CoP
                "Cannot add Expectation because it already belongs to an ExpectationSuite. "
                "If you want to update an existing Expectation, please call Expectation.save(). "
                "If you are copying this Expectation to a new ExpectationSuite, please copy "
                "it first (the core expectations and some others support copy(expectation)) "
                "and set `Expectation.id = None`."
            )
        should_save_expectation = self._has_been_saved()

        already_added = any(
            self._expectations_are_equalish(expectation, exp) for exp in self.expectations
        )
        if not already_added:
            # suite is a set-like collection, so don't add if it not unique
            if should_save_expectation:
                expectation = self._store.add_expectation(suite=self, expectation=expectation)
            self.expectations.append(expectation)

        expectation.register_save_callback(save_callback=self._save_expectation)

        self._submit_expectation_created_event(expectation=expectation)

        return expectation

    @staticmethod
    def _expectations_are_equalish(expectation_a: Expectation, expectation_b: Expectation) -> bool:
        """
        Helper method to determine if two expectations are equal enough to be considered the same.

        Note that this check is less stringent than Expectation.__eq__ and excludes a few fields
        that are not relevant for uniqueness in the suite.
        """
        exclude_params = {"id", "rendered_content", "notes", "meta"}
        # pydantic model.dict() excludes ClassVars, so we compare Expectation type explicitly
        types_are_equal = expectation_a.expectation_type == expectation_b.expectation_type
        attributes_are_equal = expectation_a.dict(exclude=exclude_params) == expectation_b.dict(
            exclude=exclude_params
        )
        return types_are_equal and attributes_are_equal

    def _submit_expectation_created_event(self, expectation: Expectation) -> None:
        if expectation.__module__.startswith("great_expectations."):
            custom_exp_type = False
            expectation_type = expectation.expectation_type
        else:
            custom_exp_type = True
            expectation_type = anonymize(expectation.expectation_type)

        submit_event(
            event=ExpectationSuiteExpectationCreatedEvent(
                expectation_id=expectation.id,
                expectation_suite_id=self.id,
                expectation_type=expectation_type,
                custom_exp_type=custom_exp_type,
            )
        )

    def _process_expectation(
        self, expectation_like: Union[Expectation, ExpectationConfiguration, dict]
    ) -> Expectation:
        """Transform an Expectation from one of its various serialized forms to the Expectation type,
        and bind it to this ExpectationSuite.

        Raises:
            ValueError: If expectation_like is of type Expectation and expectation_like.id is not None.
        """  # noqa: E501 # FIXME CoP
        from great_expectations.expectations.expectation import Expectation
        from great_expectations.expectations.expectation_configuration import (
            ExpectationConfiguration,
        )

        if isinstance(expectation_like, Expectation):
            if expectation_like.id:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "Expectations in parameter `expectations` must not belong to another ExpectationSuite. "  # noqa: E501 # FIXME CoP
                    "Instead, please use copies of Expectations, by calling `copy.copy(expectation)`."  # noqa: E501 # FIXME CoP
                )
            expectation_like.register_save_callback(save_callback=self._save_expectation)
            return expectation_like
        elif isinstance(expectation_like, ExpectationConfiguration):
            return self._build_expectation(expectation_configuration=expectation_like)
        elif isinstance(expectation_like, dict):
            return self._build_expectation(
                expectation_configuration=ExpectationConfiguration(**expectation_like)
            )
        else:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"Expected Expectation, ExpectationConfiguration, or dict, but received type {type(expectation_like)}."  # noqa: E501 # FIXME CoP
            )

    @public_api
    def delete_expectation(self, expectation: Expectation) -> Expectation:
        """Delete an Expectation from the collection.

        Example:
            >>> suite.delete_expectation(suite.expectations[0])

        Raises:
            KeyError: Expectation not found in suite.
        """
        remaining_expectations = [
            exp
            for exp in self.expectations
            if not self._expectations_are_equalish(exp, expectation)
        ]
        if len(remaining_expectations) != len(self.expectations) - 1:
            raise KeyError("No matching expectation was found.")  # noqa: TRY003 # FIXME CoP

        self.expectations = remaining_expectations

        if self._has_been_saved():
            # only persist on delete if the suite has already been saved
            try:
                self._store.delete_expectation(suite=self, expectation=expectation)
            except Exception as exc:
                # rollback this change
                # expectation suite is set-like so order of expectations doesn't matter
                self.expectations.append(expectation)
                raise exc  # noqa: TRY201 # FIXME CoP

        submit_event(
            event=ExpectationSuiteExpectationDeletedEvent(
                expectation_id=expectation.id,
                expectation_suite_id=self.id,
            )
        )

        return expectation

    @public_api
    def save(self) -> None:
        """Save this ExpectationSuite."""
        # TODO: Need to emit an event from here - we've opted out of an ExpectationSuiteUpdated event for now  # noqa: E501 # FIXME CoP
        if self._include_rendered_content:
            self.render()
        key = self._store.get_key(name=self.name, id=self.id)
        self._store.update(key=key, value=self)

    def is_fresh(self) -> ExpectationSuiteFreshnessDiagnostics:
        diagnostics = self._is_added()
        if not diagnostics.success:
            return diagnostics
        return self._is_fresh()

    def _is_added(self) -> ExpectationSuiteFreshnessDiagnostics:
        return ExpectationSuiteFreshnessDiagnostics(
            errors=[] if self.id else [ExpectationSuiteNotAddedError(name=self.name)]
        )

    def _is_fresh(self) -> ExpectationSuiteFreshnessDiagnostics:
        suite_dict: dict | None
        try:
            key = self._store.get_key(name=self.name, id=self.id)
            suite_dict = self._store.get(key=key)
        except (
            StoreBackendError,  # Generic error from stores
            InvalidKeyError,  # Ephemeral context error
        ):
            suite_dict = None
        if not suite_dict:
            return ExpectationSuiteFreshnessDiagnostics(
                errors=[ExpectationSuiteNotFoundError(name=self.name)]
            )

        suite: ExpectationSuite | None
        try:
            suite = self._store.deserialize_suite_dict(suite_dict=suite_dict)
        except PydanticValidationError:
            suite = None
        if not suite:
            return ExpectationSuiteFreshnessDiagnostics(
                errors=[ExpectationSuiteError(f"Could not deserialize suite '{self.name}'")]
            )

        return ExpectationSuiteFreshnessDiagnostics(
            errors=[] if self == suite else [ExpectationSuiteNotFreshError(name=self.name)]
        )

    def _has_been_saved(self) -> bool:
        """Has this ExpectationSuite been persisted to a Store?"""
        # todo: this should only check local keys instead of potentially querying the remote backend
        key = self._store.get_key(name=self.name, id=self.id)
        return self._store.has_key(key=key)

    def _save_expectation(self, expectation) -> Expectation:
        expectation = self._store.update_expectation(suite=self, expectation=expectation)
        submit_event(
            event=ExpectationSuiteExpectationUpdatedEvent(
                expectation_id=expectation.id, expectation_suite_id=self.id
            )
        )
        return expectation

    @property
    def expectation_configurations(self) -> list[ExpectationConfiguration]:
        return [exp.configuration for exp in self.expectations]

    @expectation_configurations.setter
    def expectation_configurations(self, value):
        raise AttributeError(  # noqa: TRY003 # FIXME CoP
            "Cannot set ExpectationSuite.expectation_configurations. "
            "Please use ExpectationSuite.expectations instead."
        )

    @override
    def __eq__(self, other):
        """ExpectationSuite equality ignores instance identity, relying only on properties."""
        if not isinstance(other, self.__class__):
            # Delegate comparison to the other instance's __eq__.
            return NotImplemented
        return all(
            (
                self.name == other.name,
                sorted(self.expectations) == sorted(other.expectations),
                self.suite_parameters == other.suite_parameters,
                self.meta == other.meta,
            )
        )

    @override
    def __hash__(self) -> int:
        return hash(
            (
                self.name,
                tuple(sorted(hash(exp) for exp in self.expectations)),
                tuple(sorted(self.suite_parameters.items())) if self.suite_parameters else (),
                tuple(sorted(self.meta.items())) if self.meta else (),
            )
        )

    def __ne__(self, other):  # type: ignore[explicit-override] # FIXME
        # By using the == operator, the returned NotImplemented is handled correctly.
        return not self == other

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return json.dumps(self.to_json_dict(), indent=2)

    @override
    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    def __deepcopy__(self, memo: dict):
        cls = self.__class__
        result = cls.__new__(cls)

        memo[id(self)] = result

        attributes_to_copy = set(ExpectationSuiteSchema().fields.keys())
        for key in attributes_to_copy:
            setattr(result, key, deepcopy(getattr(self, key), memo))

        return result

    @public_api
    @override
    def to_json_dict(self) -> Dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this ExpectationSuite.

        Returns:
            A JSON-serializable dict representation of this ExpectationSuite.
        """
        myself = expectationSuiteSchema.dump(self)
        # NOTE - JPC - 20191031: migrate to expectation-specific schemas that subclass result with properly-typed  # noqa: E501 # FIXME CoP
        # schemas to get serialization all-the-way down via dump
        expectation_configurations = [exp.configuration for exp in self.expectations]
        myself["expectations"] = convert_to_json_serializable(expectation_configurations)
        try:
            myself["suite_parameters"] = convert_to_json_serializable(myself["suite_parameters"])
        except KeyError:
            pass  # Allow suite parameters to be missing if empty
        myself["meta"] = convert_to_json_serializable(myself["meta"])
        return myself

    def remove_expectation(
        self,
        expectation_configuration: Optional[ExpectationConfiguration] = None,
        match_type: str = "domain",
        remove_multiple_matches: bool = False,
        id: Optional[Union[str, uuid.UUID]] = None,
    ) -> List[ExpectationConfiguration]:
        """Remove an ExpectationConfiguration from the ExpectationSuite.

        Args:
            expectation_configuration: A potentially incomplete (partial) Expectation Configuration to match against.
            match_type: This determines what kwargs to use when matching. Options are 'domain' to match based
                on the data evaluated by that expectation, 'success' to match based on all configuration parameters
                that influence whether an expectation succeeds based on a given batch of data, and 'runtime' to match
                based on all configuration parameters.
            remove_multiple_matches: If True, will remove multiple matching expectations.
            id: Great Expectations Cloud id for an Expectation.

        Returns:
            The list of deleted ExpectationConfigurations.

        Raises:
            TypeError: Must provide either expectation_configuration or id.
            ValueError: No match or multiple matches found (and remove_multiple_matches=False).
        """  # noqa: E501 # FIXME CoP
        expectation_configurations = [exp.configuration for exp in self.expectations]
        if expectation_configuration is None and id is None:
            raise TypeError("Must provide either expectation_configuration or id")  # noqa: TRY003 # FIXME CoP

        found_expectation_indexes = self._find_expectation_indexes(
            expectation_configuration=expectation_configuration,
            match_type=match_type,
            id=id,  # type: ignore[arg-type] # FIXME CoP
        )
        if len(found_expectation_indexes) < 1:
            raise ValueError("No matching expectation was found.")  # noqa: TRY003 # FIXME CoP

        elif len(found_expectation_indexes) > 1:
            if remove_multiple_matches:
                removed_expectations = []
                for index in sorted(found_expectation_indexes, reverse=True):
                    removed_expectations.append(expectation_configurations.pop(index))
                self.expectations = [
                    self._build_expectation(config) for config in expectation_configurations
                ]
                return removed_expectations
            else:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "More than one matching expectation was found. Specify more precise matching criteria,"  # noqa: E501 # FIXME CoP
                    "or set remove_multiple_matches=True"
                )

        else:
            result = [expectation_configurations.pop(found_expectation_indexes[0])]
            self.expectations = [
                self._build_expectation(config) for config in expectation_configurations
            ]
            return result

    def _find_expectation_indexes(
        self,
        expectation_configuration: Optional[ExpectationConfiguration] = None,
        match_type: str = "domain",
        id: Optional[str] = None,
    ) -> List[int]:
        """
        Find indexes of Expectations matching the given ExpectationConfiguration on the given match_type.
        If a id is provided, match_type is ignored and only indexes of Expectations
        with matching id are returned.

        Args:
            expectation_configuration: A potentially incomplete (partial) Expectation Configuration to match against to
                find the index of any matching Expectation Configurations on the suite.
            match_type: This determines what kwargs to use when matching. Options are 'domain' to match based
                on the data evaluated by that expectation, 'success' to match based on all configuration parameters
                 that influence whether an expectation succeeds based on a given batch of data, and 'runtime' to match
                 based on all configuration parameters
            id: Great Expectations Cloud id

        Returns: A list of indexes of matching ExpectationConfiguration

        Raises:
            InvalidExpectationConfigurationError

        """  # noqa: E501 # FIXME CoP
        from great_expectations.expectations.expectation_configuration import (
            ExpectationConfiguration,
        )

        if expectation_configuration is None and id is None:
            raise TypeError("Must provide either expectation_configuration or id")  # noqa: TRY003 # FIXME CoP

        if expectation_configuration and not isinstance(
            expectation_configuration, ExpectationConfiguration
        ):
            raise gx_exceptions.InvalidExpectationConfigurationError(  # noqa: TRY003 # FIXME CoP
                "Ensure that expectation configuration is valid."
            )

        match_indexes = []
        for idx, expectation in enumerate(self.expectations):
            if id is not None:
                if expectation.id == id:
                    match_indexes.append(idx)
            else:  # noqa: PLR5501 # FIXME CoP
                if expectation.configuration.isEquivalentTo(
                    other=expectation_configuration,  # type: ignore[arg-type] # FIXME CoP
                    match_type=match_type,
                ):
                    match_indexes.append(idx)

        return match_indexes

    def _add_expectation(
        self,
        expectation_configuration: ExpectationConfiguration,
        match_type: str = "domain",
        overwrite_existing: bool = True,
    ) -> ExpectationConfiguration:
        """
        If successful, upserts ExpectationConfiguration into this ExpectationSuite.

        Args:
            expectation_configuration: The ExpectationConfiguration to add or update
            match_type: The criteria used to determine whether the Suite already has an ExpectationConfiguration
                and so whether we should add or replace.
            overwrite_existing: If the expectation already exists, this will overwrite if True and raise an error if
                False.

        Returns:
            The ExpectationConfiguration to add or replace.

        Raises:
            More than one match
            One match if overwrite_existing = False
        """  # noqa: E501 # FIXME CoP

        found_expectation_indexes = self._find_expectation_indexes(
            expectation_configuration=expectation_configuration, match_type=match_type
        )

        if len(found_expectation_indexes) > 1:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "More than one matching expectation was found. Please be more specific with your search "  # noqa: E501 # FIXME CoP
                "criteria"
            )
        elif len(found_expectation_indexes) == 1:
            # Currently, we completely replace the expectation_configuration, but we could potentially use patch_expectation  # noqa: E501 # FIXME CoP
            # to update instead. We need to consider how to handle meta in that situation.
            # patch_expectation = jsonpatch.make_patch(self.expectations[found_expectation_index] \
            #   .kwargs, expectation_configuration.kwargs)
            # patch_expectation.apply(self.expectations[found_expectation_index].kwargs, in_place=True)  # noqa: E501 # FIXME CoP
            if overwrite_existing:
                # if existing Expectation has a id, add it back to the new Expectation Configuration
                existing_expectation_id = self.expectations[found_expectation_indexes[0]].id
                if existing_expectation_id is not None:
                    expectation_configuration.id = existing_expectation_id

                self.expectations[found_expectation_indexes[0]] = self._build_expectation(
                    expectation_configuration=expectation_configuration
                )
            else:
                raise gx_exceptions.DataContextError(  # noqa: TRY003 # FIXME CoP
                    "A matching ExpectationConfiguration already exists. If you would like to overwrite this "  # noqa: E501 # FIXME CoP
                    "ExpectationConfiguration, set overwrite_existing=True"
                )
        else:
            self.expectations.append(
                self._build_expectation(expectation_configuration=expectation_configuration)
            )

        return expectation_configuration

    def add_expectation_configurations(
        self,
        expectation_configurations: List[ExpectationConfiguration],
        match_type: str = "domain",
        overwrite_existing: bool = True,
    ) -> List[ExpectationConfiguration]:
        """Upsert a list of ExpectationConfigurations into this ExpectationSuite.

        Args:
            expectation_configurations: The List of candidate new/modifed "ExpectationConfiguration" objects for Suite.
            match_type: The criteria used to determine whether the Suite already has an "ExpectationConfiguration"
                object, matching the specified criteria, and thus whether we should add or replace (i.e., "upsert").
            overwrite_existing: If "ExpectationConfiguration" already exists, this will cause it to be overwritten if
                True and raise an error if False.

        Returns:
            The List of "ExpectationConfiguration" objects attempted to be added or replaced (can differ from the list
            of "ExpectationConfiguration" objects in "self.expectations" at the completion of this method's execution).

        Raises:
            More than one match
            One match if overwrite_existing = False
        """  # noqa: E501 # FIXME CoP
        expectation_configuration: ExpectationConfiguration
        expectation_configurations_attempted_to_be_added: List[ExpectationConfiguration] = [
            self.add_expectation_configuration(
                expectation_configuration=expectation_configuration,
                match_type=match_type,
                overwrite_existing=overwrite_existing,
            )
            for expectation_configuration in expectation_configurations
        ]
        return expectation_configurations_attempted_to_be_added

    def add_expectation_configuration(
        self,
        expectation_configuration: ExpectationConfiguration,
        match_type: str = "domain",
        overwrite_existing: bool = True,
    ) -> ExpectationConfiguration:
        """Upsert specified ExpectationConfiguration into this ExpectationSuite.

        Args:
            expectation_configuration: The ExpectationConfiguration to add or update.
            match_type: The criteria used to determine whether the Suite already has an ExpectationConfiguration
                and so whether we should add or replace.
            overwrite_existing: If the expectation already exists, this will overwrite if True and raise an error if
                False.

        Returns:
            The ExpectationConfiguration to add or replace.

        Raises:
            ValueError: More than one match
            DataContextError: One match if overwrite_existing = False

        # noqa: DAR402 # FIXME CoP
        """  # noqa: E501 # FIXME CoP
        self._build_expectation(expectation_configuration)
        return self._add_expectation(
            expectation_configuration=expectation_configuration,
            match_type=match_type,
            overwrite_existing=overwrite_existing,
        )

    def _build_expectation(
        self, expectation_configuration: ExpectationConfiguration
    ) -> Expectation:
        try:
            expectation = expectation_configuration.to_domain_obj()
            expectation.register_save_callback(save_callback=self._save_expectation)
            return expectation
        except (
            gx_exceptions.ExpectationNotFoundError,
            gx_exceptions.InvalidExpectationConfigurationError,
        ) as e:
            raise gx_exceptions.InvalidExpectationConfigurationError(  # noqa: TRY003 # FIXME CoP
                f"Could not add expectation; provided configuration is not valid: {e.message}"
            ) from e

    def render(self) -> None:
        """
        Renders content using the atomic prescriptive renderer for each expectation configuration associated with
           this ExpectationSuite to ExpectationConfiguration.rendered_content.
        """  # noqa: E501 # FIXME CoP
        for expectation in self.expectations:
            expectation.render()

    def identifier_bundle(self) -> _IdentifierBundle:
        # Utilized as a custom json_encoder
        diagnostics = self.is_fresh()
        diagnostics.raise_for_error()

        return _IdentifierBundle(name=self.name, id=self.id)


_TExpectationSuite = TypeVar("_TExpectationSuite", ExpectationSuite, dict)


class ExpectationSuiteSchema(Schema):
    name = fields.Str()
    id = fields.UUID(required=False, allow_none=True)
    expectations = fields.List(fields.Nested("ExpectationConfigurationSchema"))
    suite_parameters = fields.Dict(allow_none=True)
    meta = fields.Dict()
    notes = fields.Raw(required=False, allow_none=True)

    # NOTE: 20191107 - JPC - we may want to remove clean_empty and update tests to require the other fields;  # noqa: E501 # FIXME CoP
    # doing so could also allow us not to have to make a copy of data in the pre_dump method.
    # noinspection PyMethodMayBeStatic
    def clean_empty(self, data: _TExpectationSuite) -> _TExpectationSuite:
        if isinstance(data, ExpectationSuite):
            # We are hitting this TypeVar narrowing mypy bug: https://github.com/python/mypy/issues/10817
            data = self._clean_empty_suite(data)
        elif isinstance(data, dict):
            data = self._clean_empty_dict(data)
        return data

    @staticmethod
    def _clean_empty_suite(data: ExpectationSuite) -> ExpectationSuite:
        if not hasattr(data, "suite_parameters"):
            pass
        elif len(data.suite_parameters) == 0:
            del data.suite_parameters

        if not hasattr(data, "meta") or (data.meta is None or data.meta == []):
            pass
        elif len(data.meta) == 0:
            del data.meta
        return data

    @staticmethod
    def _clean_empty_dict(data: dict) -> dict:
        if "suite_parameters" in data and len(data["suite_parameters"]) == 0:
            data.pop("suite_parameters")
        if "meta" in data and len(data["meta"]) == 0:
            data.pop("meta")
        if "notes" in data and not data.get("notes"):
            data.pop("notes")
        return data

    # noinspection PyUnusedLocal
    @pre_dump
    def prepare_dump(self, data, **kwargs):
        data = deepcopy(data)
        for key in data:
            if key.startswith("_"):
                continue
            data[key] = convert_to_json_serializable(data[key])

        data = self.clean_empty(data)
        return data

    @post_dump(pass_original=True)
    def insert_expectations(self, data, original_data, **kwargs) -> dict:
        if isinstance(original_data, dict):
            expectations = original_data.get("expectations", [])
        else:
            expectations = original_data.expectation_configurations
        data["expectations"] = convert_to_json_serializable(expectations)
        return data

    @post_load
    def _convert_uuids_to_str(self, data, **kwargs):
        """
        Utilize UUID for data validation but convert to string before usage in business logic
        """
        attr = "id"
        uuid_val = data.get(attr)
        if uuid_val:
            data[attr] = str(uuid_val)
        return data


expectationSuiteSchema: ExpectationSuiteSchema = ExpectationSuiteSchema()



from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset

class CSVAsset(FileDataAsset): ...
class ExcelAsset(FileDataAsset): ...
class JSONAsset(FileDataAsset): ...
class ORCAsset(FileDataAsset): ...
class ParquetAsset(FileDataAsset): ...
class FeatherAsset(FileDataAsset): ...
class FWFAsset(FileDataAsset): ...
class HDFAsset(FileDataAsset): ...
class HTMLAsset(FileDataAsset): ...
class PickleAsset(FileDataAsset): ...
class SASAsset(FileDataAsset): ...
class SPSSAsset(FileDataAsset): ...
class StataAsset(FileDataAsset): ...
class XMLAsset(FileDataAsset): ...



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.expectations.core.expect_column_quantile_values_to_be_between import (
    QuantileRange,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config.big_query import BigQueryDatasourceTestConfig

COL_NAME = "my_col"

DATA = pd.DataFrame({COL_NAME: [1, 2, 2, 3, 3, 3, 4]})

ALL_DATA_SOURCES_EXCEPT_BIGQUERY = [
    ds for ds in ALL_DATA_SOURCES if not isinstance(ds, BigQueryDatasourceTestConfig)
]

# TODO: Consider more test cases before removing expect_column_quantile_values_to_be_between.json


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES_EXCEPT_BIGQUERY, data=DATA
)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnQuantileValuesToBeBetween(
        column=COL_NAME,
        quantile_ranges=QuantileRange(
            quantiles=[0, 0.333, 0.667, 1],
            value_ranges=[[0, 1], [2, 3], [3, 4], [4, 5]],
        ),
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "observed_value": {
            "quantiles": [0.0, 0.333, 0.667, 1.0],
            "values": [1, 2, 3, 4],
        },
        "details": {
            "success_details": [True, True, True, True],
        },
    }


@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_allows_unspecified_extremes(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnQuantileValuesToBeBetween(
        column=COL_NAME,
        quantile_ranges=QuantileRange(
            quantiles=[0, 0.333, 0.667, 1],
            value_ranges=[[None, 1], [2, 3], [3, 4], [4, None]],
        ),
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnQuantileValuesToBeBetween(
        column=COL_NAME,
        quantile_ranges=QuantileRange(
            quantiles=[0, 0.333, 0.667, 1],
            value_ranges=[[0, 1], [1, 2], [1, 2], [2, 3]],
        ),
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(
            {
                "quantiles": [0, 0.333, 0.667, 1],
                "value_ranges": [[0, 1], [2, 3], [3, 4], [4, 5]],
            },
            True,
            id="success",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_quantile_ranges_(
    batch_for_datasource: Batch, suite_param_value: dict, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_quantile_values_to_be_between"
    expectation = gxe.ExpectColumnQuantileValuesToBeBetween(
        column=COL_NAME,
        quantile_ranges={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(False, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_allow_relative_error_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_quantile_values_to_be_between"
    expectation = gxe.ExpectColumnQuantileValuesToBeBetween(
        column=COL_NAME,
        quantile_ranges=QuantileRange(
            quantiles=[0, 0.333, 0.667, 1],
            value_ranges=[[0, 1], [2, 3], [3, 4], [4, 5]],
        ),
        allow_relative_error={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



import copy
from unittest import mock

import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.data_context import EphemeralDataContext


@pytest.fixture
def new_site_config() -> dict:
    return {
        "class_name": "SiteBuilder",
        "module_name": "great_expectations.render.renderer.site_builder",
        "store_backend": {
            "module_name": "great_expectations.data_context.store.tuple_store_backend",
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": "/my_new_site/",
        },
        "site_index_builder": {"class_name": "DefaultSiteIndexBuilder"},
    }


class TestAddDataDocsSite:
    @pytest.mark.unit
    def test_add_data_docs_site(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
        new_site_config: dict,
    ):
        # Add a new site
        new_site_name = "my_new_site"
        ephemeral_context_with_defaults.add_data_docs_site(
            site_name=new_site_name, site_config=new_site_config
        )

        # Check that the new site is present
        assert new_site_name in ephemeral_context_with_defaults.get_site_names()

    @pytest.mark.unit
    def test_add_data_docs_site_persists(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
        new_site_config: dict,
    ):
        new_site_name = "my_new_site"
        with mock.patch(
            "great_expectations.data_context.EphemeralDataContext._save_project_config"
        ) as mock_save_project_config:
            ephemeral_context_with_defaults.add_data_docs_site(
                site_name=new_site_name, site_config=new_site_config
            )

        mock_save_project_config.assert_called_once()

    @pytest.mark.unit
    def test_add_data_docs_site_already_existing_site_raises_exception(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
        new_site_config: dict,
    ):
        # Check fixture configuration
        existing_site_name = "local_site"
        assert existing_site_name in ephemeral_context_with_defaults.get_site_names()

        with pytest.raises(gx_exceptions.InvalidKeyError) as e:
            new_site_name = existing_site_name
            ephemeral_context_with_defaults.add_data_docs_site(
                site_name=new_site_name, site_config=new_site_config
            )

        assert "Data Docs Site `local_site` already exists in the Data Context." in str(e.value)


class TestListDataDocsSites:
    @pytest.mark.unit
    def test_list_data_docs_sites(self, ephemeral_context_with_defaults: EphemeralDataContext):
        site_names = [d for d in ephemeral_context_with_defaults.list_data_docs_sites()]
        assert site_names == ["local_site"]


class TestUpdateDataDocsSite:
    @pytest.mark.unit
    def test_update_data_docs_site(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
        new_site_config: dict,
    ):
        # Add a new site
        new_site_name = "my_new_site"
        ephemeral_context_with_defaults.add_data_docs_site(
            site_name=new_site_name, site_config=new_site_config
        )

        # Update the new site
        updated_site_config = copy.deepcopy(new_site_config)
        updated_site_config["store_backend"]["base_directory"] = "/my_updated_site/"

        ephemeral_context_with_defaults.update_data_docs_site(new_site_name, updated_site_config)

        # Check the updated site config
        sites = ephemeral_context_with_defaults.variables.data_docs_sites
        assert sites[new_site_name]["store_backend"]["base_directory"] == "/my_updated_site/"

    @pytest.mark.unit
    def test_update_data_docs_site_persists(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
        new_site_config: dict,
    ):
        # Add a new site
        new_site_name = "my_new_site"
        ephemeral_context_with_defaults.add_data_docs_site(
            site_name=new_site_name, site_config=new_site_config
        )

        # Update the new site
        updated_site_config = copy.deepcopy(new_site_config)
        updated_site_config["store_backend"]["base_directory"] = "/my_updated_site/"

        with mock.patch(
            "great_expectations.data_context.EphemeralDataContext._save_project_config"
        ) as mock_save_project_config:
            ephemeral_context_with_defaults.update_data_docs_site(
                new_site_name, updated_site_config
            )

        mock_save_project_config.assert_called_once()

    @pytest.mark.unit
    def test_update_data_docs_site_missing_site_raises_exception(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
        new_site_config: dict,
    ):
        # Check fixture configuration
        assert "missing" not in ephemeral_context_with_defaults.get_site_names()

        with pytest.raises(gx_exceptions.InvalidKeyError) as e:
            ephemeral_context_with_defaults.update_data_docs_site(
                site_name="missing", site_config=new_site_config
            )

        assert "Data Docs Site `missing` does not already exist in the Data Context." in str(
            e.value
        )


class TestDeleteDataDocsSite:
    @pytest.mark.unit
    def test_delete_data_docs_site(self, ephemeral_context_with_defaults: EphemeralDataContext):
        # Check fixture configuration
        existing_site_name = "local_site"
        assert existing_site_name in ephemeral_context_with_defaults.get_site_names()

        ephemeral_context_with_defaults.delete_data_docs_site(existing_site_name)

        # Check that the site is no longer present
        assert existing_site_name not in ephemeral_context_with_defaults.get_site_names()

    @pytest.mark.unit
    def test_delete_data_docs_site_persists(
        self, ephemeral_context_with_defaults: EphemeralDataContext
    ):
        # Check fixture configuration
        existing_site_name = "local_site"
        assert existing_site_name in ephemeral_context_with_defaults.get_site_names()

        with mock.patch(
            "great_expectations.data_context.EphemeralDataContext._save_project_config"
        ) as mock_save_project_config:
            ephemeral_context_with_defaults.delete_data_docs_site(existing_site_name)

        mock_save_project_config.assert_called_once()

    @pytest.mark.unit
    def test_delete_data_docs_site_missing_site_raises_exception(
        self,
        ephemeral_context_with_defaults: EphemeralDataContext,
    ):
        # Check fixture configuration
        assert "missing" not in ephemeral_context_with_defaults.get_site_names()

        with pytest.raises(gx_exceptions.InvalidKeyError) as e:
            ephemeral_context_with_defaults.delete_data_docs_site("missing")

        assert "Data Docs Site `missing` does not already exist in the Data Context." in str(
            e.value
        )



from __future__ import annotations

import abc
import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional

from great_expectations.core.profiler_types_mapping import ProfilerTypeMapping

if TYPE_CHECKING:
    from great_expectations.core.expectation_suite import ExpectationSuite

logger = logging.getLogger(__name__)


class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class OrderedProfilerCardinality(OrderedEnum):
    NONE = 0
    ONE = 1
    TWO = 2
    VERY_FEW = 3
    FEW = 4
    MANY = 5
    VERY_MANY = 6
    UNIQUE = 7

    @classmethod
    def get_basic_column_cardinality(cls, num_unique=0, pct_unique=0) -> OrderedProfilerCardinality:
        """
        Takes the number and percentage of unique values in a column and returns the column cardinality.
        If you are unexpectedly returning a cardinality of "None", ensure that you are passing in values for both
        num_unique and pct_unique.

        Args:
            num_unique: The number of unique values in a column
            pct_unique: The percentage of unique values in a column

        Returns:
            The column cardinality
        """  # noqa: E501 # FIXME CoP
        if pct_unique == 1.0:
            cardinality = cls.UNIQUE
        elif num_unique == 1:
            cardinality = cls.ONE
        elif num_unique == 2:  # noqa: PLR2004 # FIXME CoP
            cardinality = cls.TWO
        elif 0 < num_unique < 20:  # noqa: PLR2004 # FIXME CoP
            cardinality = cls.VERY_FEW
        elif 0 < num_unique < 60:  # noqa: PLR2004 # FIXME CoP
            cardinality = cls.FEW
        elif num_unique is None or num_unique == 0 or pct_unique is None:
            cardinality = cls.NONE
        elif pct_unique > 0.1:  # noqa: PLR2004 # FIXME CoP
            cardinality = cls.VERY_MANY
        else:
            cardinality = cls.MANY
        return cardinality


class ProfilerDataType(Enum):
    """Useful data types for building profilers."""

    INT = "int"
    FLOAT = "float"
    NUMERIC = "numeric"
    STRING = "string"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    UNKNOWN = "unknown"


class ProfilerCardinality(Enum):
    """Useful cardinality categories for building profilers."""

    NONE = "none"
    ONE = "one"
    TWO = "two"
    FEW = "few"
    VERY_FEW = "very few"
    MANY = "many"
    VERY_MANY = "very many"
    UNIQUE = "unique"


profiler_data_types_with_mapping = {
    "INT": list(ProfilerTypeMapping.INT_TYPE_NAMES),
    "FLOAT": list(ProfilerTypeMapping.FLOAT_TYPE_NAMES),
    "NUMERIC": (
        list(ProfilerTypeMapping.INT_TYPE_NAMES) + list(ProfilerTypeMapping.FLOAT_TYPE_NAMES)
    ),
    "STRING": list(ProfilerTypeMapping.STRING_TYPE_NAMES),
    "BOOLEAN": list(ProfilerTypeMapping.BOOLEAN_TYPE_NAMES),
    "DATETIME": list(ProfilerTypeMapping.DATETIME_TYPE_NAMES),
    "UNKNOWN": ["unknown"],
}


class ProfilerSemanticTypes(Enum):
    DATETIME = "DATETIME"
    NUMERIC = "NUMERIC"
    STRING = "STRING"
    VALUE_SET = "VALUE_SET"
    BOOLEAN = "BOOLEAN"
    OTHER = "OTHER"


class Profiler(metaclass=abc.ABCMeta):
    """
    Profiler creates suites from various sources of truth.

    These sources of truth can be data or non-data sources such as DDLs.

    When implementing a Profiler ensure that you:
    - Implement a . _profile() method
    - Optionally implement .validate() method that verifies you are running on the right
      kind of object. You should raise an appropriate Exception if the object is not valid.
    """

    def __init__(self, configuration: Optional[dict] = None) -> None:
        self.configuration = configuration

    def validate(  # noqa: B027  # empty-method-without-abstract-decorator
        self, item_to_validate: Any
    ) -> None:
        """Raise an exception if `item_to_validate` cannot be profiled.

        Args:
            item_to_validate: The item to validate.
        """
        pass

    def profile(self, item_to_profile: Any, suite_name: Optional[str] = None) -> ExpectationSuite:
        self.validate(item_to_profile)
        expectation_suite = self._profile(item_to_profile, suite_name=suite_name)
        return expectation_suite

    @abc.abstractmethod
    def _profile(self, item_to_profile: Any, suite_name: Optional[str] = None) -> ExpectationSuite:
        pass



import datetime
import re

import numpy as np
import pytest

from great_expectations.util import convert_to_json_serializable, ensure_json_serializable

try:
    from shapely.geometry import LineString, MultiPolygon, Point, Polygon
except ImportError:
    Point = None
    LineString = None
    Polygon = None
    MultiPolygon = None


@pytest.mark.unit
@pytest.mark.skipif(Polygon is None, reason="Requires shapely.geometry.Polygon")
def test_serialization_of_shapely_polygon():
    polygon = Polygon([(0, 0), (2, 0), (2, 2), (0, 2)])
    assert convert_to_json_serializable(polygon) == "POLYGON ((0 0, 2 0, 2 2, 0 2, 0 0))"


@pytest.mark.unit
@pytest.mark.skipif(
    any([MultiPolygon is None, Polygon is None]),
    reason="Requires shapely.geometry.Polygon and MultiPolygon",
)
def test_serialization_of_shapely_multipolygon():
    multi_polygon = MultiPolygon([Polygon([(1, 2), (3, 4), (5, 6)])])
    assert convert_to_json_serializable(multi_polygon) == "MULTIPOLYGON (((1 2, 3 4, 5 6, 1 2)))"


@pytest.mark.unit
@pytest.mark.skipif(Point is None, reason="Requires shapely.geometry.Point")
def test_serialization_of_shapely_point():
    point = Point(1, 2)
    assert convert_to_json_serializable(point) == "POINT (1 2)"


@pytest.mark.unit
@pytest.mark.skipif(LineString is None, reason="Requires shapely.geometry.LineString")
def test_serialization_of_shapely_linestring():
    point = LineString([(0, 0), (1, 1), (1, -1)])
    assert convert_to_json_serializable(point) == "LINESTRING (0 0, 1 1, 1 -1)"


@pytest.mark.unit
def test_serialization_of_bytes():
    data = b"\xc0\xa8\x00\x01"
    assert convert_to_json_serializable(data) == "b'\\xc0\\xa8\\x00\\x01'"


@pytest.mark.unit
def test_serialization_numpy_datetime():
    datetime_to_test = "2022-12-08T12:56:23.423"
    data = np.datetime64(datetime_to_test)
    assert convert_to_json_serializable(data) == datetime_to_test


@pytest.mark.unit
def test_serialization_of_pattern():
    pattern_to_test = r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv"
    data = re.compile(pattern_to_test)
    assert convert_to_json_serializable(data) == pattern_to_test


@pytest.mark.unit
@pytest.mark.parametrize(
    "data", [pytest.param({"t": datetime.time(hour=1, minute=30, second=45)}, id="datetime.time")]
)
def test_convert_to_json_serializable_converts_correctly(data: dict):
    ret = convert_to_json_serializable(data)
    assert ret == {"t": "01:30:45"}


@pytest.mark.unit
@pytest.mark.parametrize(
    "data", [pytest.param({"t": datetime.time(hour=1, minute=30, second=45)}, id="datetime.time")]
)
def test_ensure_json_serializable(data: dict):
    ensure_json_serializable(data)
    # Passes if no exception raised



import logging

import pytest

from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent import AlloyDatasource
from great_expectations.execution_engine import SqlAlchemyExecutionEngine

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def mock_test_connection(monkeypatch: pytest.MonkeyPatch):
    """Patches the test_connection method of the AlloyDatasource class to return True."""

    def _mock_test_connection(self: AlloyDatasource) -> bool:
        LOGGER.warning(
            f"Mocked {self.__class__.__name__}.test_connection() called and returning True"
        )
        return True

    monkeypatch.setattr(AlloyDatasource, "test_connection", _mock_test_connection)


@pytest.mark.unit
def test_add_alloy_datasource(
    mock_test_connection,
    empty_data_context: AbstractDataContext,
):
    test_datasource_name = "test_datasource"
    test_connection_string = "postgresql://username:@hostname/database_name"
    source = empty_data_context.data_sources.add_alloy(
        name=test_datasource_name, connection_string=test_connection_string
    )
    assert source.type == "alloy"
    assert source.name == test_datasource_name
    assert source.execution_engine_type is SqlAlchemyExecutionEngine
    assert source.assets == []



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

INT_COL_A = "INT_COL_A"
INT_COL_B = "INT_COL_B"
INT_COL_C = "INT_COL_C"
STRING_COL_A = "STRING_COL_A"
STRING_COL_B = "STRING_COL_B"


DATA = pd.DataFrame(
    {
        INT_COL_A: [1, 1, 2, 3],
        INT_COL_B: [2, 2, 3, 4],
        INT_COL_C: [3, 3, 4, 4],
        STRING_COL_A: ["a", "b", "c", "d"],
        STRING_COL_B: ["x", "y", "z", "a"],
    }
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_golden_path(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(
        column_list=[INT_COL_A, INT_COL_B]
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(
                column_list=[INT_COL_A, INT_COL_B, INT_COL_C], mostly=0.75
            ),
            id="mostly",
        ),
        pytest.param(
            gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(
                column_list=[STRING_COL_A, STRING_COL_B]
            ),
            id="strings_dont_error",
        ),
        pytest.param(
            gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(
                column_list=[STRING_COL_A, INT_COL_A]
            ),
            id="strings_and_ints",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(column_list=[INT_COL_B, INT_COL_C]),
            id="one_non_unique",
        ),
        pytest.param(
            gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(
                column_list=[INT_COL_A, INT_COL_B, INT_COL_C], mostly=0.8
            ),
            id="mostly_threshold_not_met",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param("all_values_are_missing", True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_ignore_row_if_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_select_column_values_to_be_unique_within_record"
    expectation = gxe.ExpectSelectColumnValuesToBeUniqueWithinRecord(
        column_list=[INT_COL_A, INT_COL_B, INT_COL_C],
        mostly=0.75,
        ignore_row_if={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from __future__ import annotations

import random
from typing import TYPE_CHECKING

import pytest
import sqlalchemy as sa
from pytest import param
from sqlalchemy.sql import text

from great_expectations.datasource.fluent import (
    SnowflakeDatasource,
    TestConnectionError,
)

if TYPE_CHECKING:
    from sqlalchemy.engine.reflection import Inspector

    from great_expectations.data_context import AbstractDataContext as DataContext


@pytest.mark.snowflake
class TestSnowflake:
    @pytest.mark.xfail(
        raises=AssertionError,
    )  # inspector.get_table_names() fails with this role
    @pytest.mark.parametrize(
        "connection_string",
        [
            param(
                "snowflake://ci:${SNOWFLAKE_CI_USER_PASSWORD}@oca29081.us-east-1/ci/public?warehouse=ci&role=ci_no_select",
                id="role wo select",
            ),
        ],
    )
    def test_un_queryable_asset_should_raise_error(
        self, context: DataContext, connection_string: str
    ):
        """
        If we try to add an asset that is not queryable with the current datasource
        connection details, then we should expect a TestConnectionError.
        https://docs.snowflake.com/en/developer-guide/python-connector/sqlalchemy#connection-parameters
        """
        snowflake_ds: SnowflakeDatasource = context.data_sources.add_snowflake(
            "my_ds", connection_string=connection_string
        )

        inspector: Inspector = sa.inspection.inspect(snowflake_ds.get_engine())
        inspector_tables: list[str] = list(inspector.get_table_names(schema="public"))
        print(f"tables: {len(inspector_tables)}\n{inspector_tables}")
        random.shuffle(inspector_tables)

        unqueryable_table: str = ""
        for table_name in inspector_tables:
            try:
                # query the asset, if it fails then we should expect a TestConnectionError
                # expect the sql ProgrammingError to be raised
                # we are only testing the failure case here
                with snowflake_ds.get_engine().connect() as conn:
                    conn.execute(text(f"SELECT * FROM {table_name} LIMIT 1;"))
                print(f"{table_name} is queryable")
            except sa.exc.ProgrammingError:
                print(f"{table_name} is not queryable")
                unqueryable_table = table_name
                break
        assert unqueryable_table, "no unqueryable tables found, cannot run test"

        with pytest.raises(TestConnectionError) as exc_info:
            asset = snowflake_ds.add_table_asset(
                name="un-reachable asset", table_name=unqueryable_table
            )
            print(f"\n  Uh oh, asset should not have been created...\n{asset!r}")
        print(f"\n  TestConnectionError was raised as expected.\n{exc_info.exconly()}")

    @pytest.mark.parametrize(
        "connection_string",
        [
            param(
                "snowflake://ci:${SNOWFLAKE_CI_USER_PASSWORD}@oca29081.us-east-1/ci/public?warehouse=ci&role=ci&database=ci&schema=public",
                id="full connection string",
            ),
        ],
    )
    def test_queryable_asset_should_pass_test_connection(
        self, context: DataContext, connection_string: str
    ):
        snowflake_ds: SnowflakeDatasource = context.data_sources.add_snowflake(
            "my_ds", connection_string=connection_string
        )

        inspector: Inspector = sa.inspection.inspect(snowflake_ds.get_engine())
        inspector_tables = list(inspector.get_table_names())
        print(f"tables: {len(inspector_tables)}\n{inspector_tables}")

        table_name = random.choice(inspector_tables)

        # query the table to make sure it is queryable
        with snowflake_ds.get_engine().connect() as conn:
            conn.execute(text(f"SELECT * FROM {table_name} LIMIT 1;"))

        # the table is queryable so the `add_table_asset()` should pass the test_connection step
        asset = snowflake_ds.add_table_asset(name="reachable asset", table_name=table_name)
        print(f"\n  Yay, asset was created!\n{asset!r}")


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Generator

import pytest

from great_expectations import ValidationDefinition
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.datasource.fluent.spark_datasource import DataFrameAsset
from tests.integration.cloud.end_to_end.conftest import construct_spark_df_from_pandas

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.checkpoint.checkpoint import Checkpoint, CheckpointResult
    from great_expectations.compatibility import pyspark
    from great_expectations.core import ExpectationSuite
    from great_expectations.data_context import CloudDataContext
    from great_expectations.datasource.fluent import (
        DataAsset,
        SparkDatasource,
    )


@pytest.fixture(scope="module")
def datasource(
    context: CloudDataContext,
    datasource_name: str,
) -> SparkDatasource:
    """Test Adding and Updating the Datasource associated with this module.
    Note: There is no need to test Get or Delete Datasource.
    Those assertions can be found in the datasource_name fixture."""
    datasource = context.data_sources.add_spark(
        name=datasource_name,
        persist=True,
    )
    datasource.persist = False
    datasource_dict = datasource.dict()
    datasource = context.data_sources.add_or_update_spark(**datasource_dict)
    assert datasource.persist is False, (
        "The datasource was not updated in the previous method call."
    )
    datasource.persist = True
    datasource_dict = datasource.dict()
    datasource = context.data_sources.add_or_update_spark(**datasource_dict)
    assert datasource.persist is True, "The datasource was not updated in the previous method call."
    return datasource


@pytest.fixture(scope="module")
def data_asset(
    datasource: SparkDatasource,
) -> Generator[DataAsset, None, None]:
    """Test the entire Data Asset CRUD lifecycle here and in Data Asset-specific fixtures."""
    asset_name = f"da_{uuid.uuid4().hex}"
    yield datasource.add_dataframe_asset(
        name=asset_name,
    )
    datasource.delete_asset(name=asset_name)
    with pytest.raises(LookupError):
        datasource.get_asset(name=asset_name)


@pytest.fixture(scope="module")
def spark_dataframe(
    spark_session: pyspark.SparkSession, test_data: pd.DataFrame
) -> pyspark.DataFrame:
    return construct_spark_df_from_pandas(spark_session, test_data)


@pytest.fixture(scope="module")
def batch_definition(
    context: CloudDataContext,
    data_asset: DataFrameAsset,
) -> BatchDefinition:
    batch_def_name = f"batch_def_{uuid.uuid4().hex}"
    return data_asset.add_batch_definition_whole_dataframe(
        name=batch_def_name,
    )


@pytest.fixture(scope="module")
def validation_definition(
    context: CloudDataContext,
    expectation_suite: ExpectationSuite,
    batch_definition: BatchDefinition,
) -> Generator[ValidationDefinition, None, None]:
    validation_def_name = f"val_def_{uuid.uuid4().hex}"
    yield context.validation_definitions.add(
        ValidationDefinition(
            name=validation_def_name,
            data=batch_definition,
            suite=expectation_suite,
        )
    )
    context.validation_definitions.delete(name=validation_def_name)


@pytest.fixture(scope="module")
def checkpoint(
    checkpoint: Checkpoint,
    validation_definition: ValidationDefinition,
) -> Checkpoint:
    checkpoint.validation_definitions = [validation_definition]
    checkpoint.save()
    return checkpoint


@pytest.mark.cloud
def test_checkpoint_run(checkpoint: Checkpoint, spark_dataframe: pyspark.DataFrame):
    """Test running a Checkpoint that was created using the entities defined in this module."""
    checkpoint_result: CheckpointResult = checkpoint.run(
        batch_parameters={
            "dataframe": spark_dataframe,
        }
    )
    assert checkpoint_result.success



from unittest import mock

import pytest

from great_expectations.data_context.util import file_relative_path
from great_expectations.exceptions.exceptions import MetricResolutionError
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from tests.expectations.test_util import get_table_columns_metric


@pytest.mark.sqlite
def test_table_column_introspection(sa, capsys):
    db_file = file_relative_path(
        __file__,
        "../../test_sets/test_cases_for_sql_data_connector.db",
    )
    eng = sa.create_engine(f"sqlite:///{db_file}", echo=True)
    engine = SqlAlchemyExecutionEngine(engine=eng)
    batch_data = SqlAlchemyBatchData(
        execution_engine=engine, table_name="table_partitioned_by_date_column__A"
    )
    engine.load_batch_data("__", batch_data)
    assert isinstance(batch_data.selectable, sa.Table)
    assert batch_data.selectable.name == "table_partitioned_by_date_column__A"
    assert batch_data.selectable.schema is None

    insp = engine.get_inspector()
    columns = insp.get_columns(batch_data.selectable.name, schema=batch_data.selectable.schema)
    assert [x["name"] for x in columns] == [
        "index",
        "id",
        "date",
        "event_type",
        "favorite_color",
    ]
    assert capsys.readouterr().out

    # Assert that using the same inspector again does not cause a query
    columns = insp.get_columns(batch_data.selectable.name, schema=batch_data.selectable.schema)
    assert [x["name"] for x in columns] == [
        "index",
        "id",
        "date",
        "event_type",
        "favorite_color",
    ]
    assert not capsys.readouterr().out

    # Add another batch
    batch_data = SqlAlchemyBatchData(
        execution_engine=engine,
        table_name="table_that_should_be_partitioned_by_random_hash__H",
    )
    engine.load_batch_data("__1", batch_data)
    assert isinstance(batch_data.selectable, sa.Table)
    assert batch_data.selectable.name == "table_that_should_be_partitioned_by_random_hash__H"
    assert batch_data.selectable.schema is None

    columns = insp.get_columns(batch_data.selectable.name, schema=batch_data.selectable.schema)
    assert [x["name"] for x in columns] == [
        "index",
        "id",
        "event_type",
        "favorite_color",
    ]

    # Should have to re-inspect due to new batch
    assert capsys.readouterr().out


@pytest.mark.sqlite
def test_table_column_type__sqlalchemy_happy_path(sa, capsys):
    db_file = file_relative_path(
        __file__,
        "../../test_sets/test_cases_for_sql_data_connector.db",
    )
    eng = sa.create_engine(f"sqlite:///{db_file}", echo=True)
    engine = SqlAlchemyExecutionEngine(engine=eng)
    with mock.patch(
        "great_expectations.execution_engine.sqlalchemy_execution_engine.sa.inspect",
        wraps=sa.inspect,
    ) as mock_inspect:
        batch_data = SqlAlchemyBatchData(
            execution_engine=engine, table_name="table_partitioned_by_date_column__A"
        )
        engine.load_batch_data("__", batch_data)
        assert isinstance(batch_data.selectable, sa.Table)
        assert batch_data.selectable.name == "table_partitioned_by_date_column__A"
        assert batch_data.selectable.schema is None

        _table_columns_metric, results = get_table_columns_metric(execution_engine=engine)

        assert results[("table.columns", (), ())] == [
            "index",
            "id",
            "date",
            "event_type",
            "favorite_color",
        ]

        assert mock_inspect.call_count == 1

        _table_columns_metric, results = get_table_columns_metric(execution_engine=engine)

        assert results[("table.columns", (), ())] == [
            "index",
            "id",
            "date",
            "event_type",
            "favorite_color",
        ]

        # Should not have to re-inspect
        assert mock_inspect.call_count == 1

        # Add another batch
        batch_data = SqlAlchemyBatchData(
            execution_engine=engine,
            table_name="table_that_should_be_partitioned_by_random_hash__H",
        )
        engine.load_batch_data("__1", batch_data)
        assert isinstance(batch_data.selectable, sa.Table)
        assert batch_data.selectable.name == "table_that_should_be_partitioned_by_random_hash__H"
        assert batch_data.selectable.schema is None

        _table_columns_metric, results = get_table_columns_metric(execution_engine=engine)

        assert results[("table.columns", (), ())] == [
            "index",
            "id",
            "event_type",
            "favorite_color",
        ]

        # Should have to re-inspect due to new batch
        # But should use same inspector
        assert mock_inspect.call_count == 1


@pytest.mark.sqlite
def test_table_column_types__sqlalchemy_table_not_found(sa):
    db_file = file_relative_path(
        __file__,
        "../../test_sets/test_cases_for_sql_data_connector.db",
    )
    eng = sa.create_engine(f"sqlite:///{db_file}", echo=True)
    engine = SqlAlchemyExecutionEngine(engine=eng)

    # Table `table_partitioned_by_date_column__B` does not exist in database
    batch_data = SqlAlchemyBatchData(
        execution_engine=engine, table_name="table_partitioned_by_date_column__B"
    )
    engine.load_batch_data("__1", batch_data)
    assert isinstance(batch_data.selectable, sa.Table)
    assert batch_data.selectable.name == "table_partitioned_by_date_column__B"
    assert batch_data.selectable.schema is None

    with pytest.raises(MetricResolutionError) as exc:
        _table_columns_metric, _results = get_table_columns_metric(execution_engine=engine)
    assert "(sqlite3.OperationalError) no such table: table_partitioned_by_date_column__B" in str(
        exc.value
    )



from datetime import datetime, timezone

import pandas as pd

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import SqliteDatasourceTestConfig

DATE_COL = "date"
VALUE_COL = "value"

LAST_YEAR = "last year"
FIRST_DAY_OF_THE_YEAR = "first day of the year"
FIRST_DAY_OF_THE_MONTH = "first day of the month"
SECOND_DAY_OF_THE_MONTH = "second day of the month"

DATA = pd.DataFrame(
    {
        DATE_COL: [
            datetime(year=2023, month=1, day=1, tzinfo=timezone.utc).date(),
            datetime(year=2024, month=1, day=1, tzinfo=timezone.utc).date(),
            datetime(year=2024, month=2, day=1, tzinfo=timezone.utc).date(),
            datetime(year=2024, month=2, day=2, tzinfo=timezone.utc).date(),
        ],
        VALUE_COL: [
            LAST_YEAR,
            FIRST_DAY_OF_THE_YEAR,
            FIRST_DAY_OF_THE_MONTH,
            SECOND_DAY_OF_THE_MONTH,
        ],
    }
)

JUST_SQLITE = [SqliteDatasourceTestConfig()]


class TestPartitioning:
    """Tests to show that we partition sqlite data sourdces correctly.

    All tests use ExpectColumnDistinctValuesToEqualSet to detect that we are just seeing the
    appropriate rows.
    """

    @parameterize_batch_for_data_sources(
        data_source_configs=JUST_SQLITE,
        data=pd.DataFrame(DATA),
    )
    def test_yearly_partitioning(self, asset_for_datasource: TableAsset) -> None:
        batch_def = asset_for_datasource.add_batch_definition_yearly("yearly", column=DATE_COL)
        batch = batch_def.get_batch()

        result = batch.validate(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=VALUE_COL,
                value_set=[
                    # NOT LAST_YEAR
                    FIRST_DAY_OF_THE_YEAR,
                    FIRST_DAY_OF_THE_MONTH,
                    SECOND_DAY_OF_THE_MONTH,
                ],
            )
        )
        assert result.success

    @parameterize_batch_for_data_sources(
        data_source_configs=JUST_SQLITE,
        data=pd.DataFrame(DATA),
    )
    def test_monthly_partitioning(self, asset_for_datasource: TableAsset) -> None:
        batch_def = asset_for_datasource.add_batch_definition_monthly("monthly", column=DATE_COL)
        batch = batch_def.get_batch()

        result = batch.validate(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=VALUE_COL,
                value_set=[
                    # NOT LAST_YEAR
                    # NOT FIRST_DAY_OF_THE_YEAR,
                    FIRST_DAY_OF_THE_MONTH,
                    SECOND_DAY_OF_THE_MONTH,
                ],
            )
        )
        assert result.success

    @parameterize_batch_for_data_sources(
        data_source_configs=JUST_SQLITE,
        data=pd.DataFrame(DATA),
    )
    def test_daily_partitioning(self, asset_for_datasource: TableAsset) -> None:
        batch_def = asset_for_datasource.add_batch_definition_daily("daily", column=DATE_COL)
        batch = batch_def.get_batch()

        result = batch.validate(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=VALUE_COL,
                value_set=[
                    # NOT LAST_YEAR
                    # NOT FIRST_DAY_OF_THE_YEAR,
                    # NOT FIRST_DAY_OF_THE_MONTH,
                    SECOND_DAY_OF_THE_MONTH,
                ],
            )
        )
        assert result.success

    @parameterize_batch_for_data_sources(
        data_source_configs=JUST_SQLITE,
        data=pd.DataFrame(DATA),
    )
    def test_order_ascending__true(self, asset_for_datasource: TableAsset) -> None:
        batch_def = asset_for_datasource.add_batch_definition_daily(
            "daily_ascending", column=DATE_COL, sort_ascending=True
        )
        batch = batch_def.get_batch()

        result = batch.validate(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=VALUE_COL,
                value_set=[
                    SECOND_DAY_OF_THE_MONTH,
                ],
            )
        )
        assert result.success

    @parameterize_batch_for_data_sources(
        data_source_configs=JUST_SQLITE,
        data=pd.DataFrame(DATA),
    )
    def test_order_ascending__false(self, asset_for_datasource: TableAsset) -> None:
        batch_def = asset_for_datasource.add_batch_definition_daily(
            "daily_descending", column=DATE_COL, sort_ascending=False
        )
        batch = batch_def.get_batch()

        result = batch.validate(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=VALUE_COL,
                value_set=[
                    LAST_YEAR,
                ],
            )
        )
        assert result.success



import pandas as pd
import pytest

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.null_count import (
    ColumnNullCount,
    ColumnNullCountResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import (
    PANDAS_DATA_SOURCES,
    SPARK_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

STRING_COLUMN_NAME = "whatevs"
DATA_FRAME = pd.DataFrame(
    {
        STRING_COLUMN_NAME: ["a", None, "c", "d", None],
    },
    dtype="object",
)


class TestColumnNullCount:
    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES + PANDAS_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success(self, batch_for_datasource: Batch) -> None:
        metric = ColumnNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnNullCountResult)
        assert metric_result.value == 2

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=DATA_FRAME,
    )
    @pytest.mark.xfail(strict=True)
    def test_spark(self, batch_for_datasource: Batch) -> None:
        metric = ColumnNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnNullCountResult)
        assert metric_result.value == 2



# setup
import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations.core.expectation_suite import ExpectationSuite

context = gx.get_context()
suite_name = "test-suite"

# create a suite
suite = context.suites.add(suite=ExpectationSuite(name=suite_name))

# get a suite
suite = context.suites.get(name=suite_name)

# add expectations
suite.add_expectation(
    gxe.ExpectColumnValuesToBeInSet(column="passenger_count", value_set=[1, 2, 3, 4, 5])
)
suite.add_expectation(gxe.ExpectColumnValuesToNotBeNull(column="pickup_datetime"))

# find a single expectation by type and/or domain
expectation = next(
    expectation
    for expectation in suite.expectations
    if isinstance(expectation, gxe.ExpectColumnValuesToNotBeNull)
    and expectation.column == "pickup_datetime"
)

# update a single expectation
expectation = next(
    expectation
    for expectation in suite.expectations
    if isinstance(expectation, gxe.ExpectColumnValuesToNotBeNull)
)
expectation.column = "pickup_location_id"
expectation.save()

# update multiple expectations as a batch
for expectation in suite.expectations:
    expectation.notes = "This Expectation was generated as part of GX Documentation."
suite.save()

# delete an expectation
expectation = next(
    expectation
    for expectation in suite.expectations
    if isinstance(expectation, gxe.ExpectColumnValuesToNotBeNull)
)
suite.delete_expectation(expectation=expectation)

# delete a suite
context.suites.delete(name=suite_name)



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

AZURE_BLOB_STORAGE_NOT_IMPORTED = NotImported(
    "azure blob storage components are not installed, please 'pip install azure-storage-blob azure-identity azure-keyvault-secrets'"  # noqa: E501 # FIXME CoP
)

try:
    from azure import storage
except ImportError:
    storage = AZURE_BLOB_STORAGE_NOT_IMPORTED

try:
    from azure.identity import DefaultAzureCredential
except ImportError:
    DefaultAzureCredential = AZURE_BLOB_STORAGE_NOT_IMPORTED  # type: ignore[misc] # assigning to type

try:
    from azure.keyvault.secrets import SecretClient
except (ImportError, AttributeError):
    SecretClient = AZURE_BLOB_STORAGE_NOT_IMPORTED  # type: ignore[misc] # assigning to type

try:
    from azure.storage.blob import ContentSettings
except (ImportError, AttributeError):
    ContentSettings = AZURE_BLOB_STORAGE_NOT_IMPORTED  # type: ignore[misc] # assigning to type
try:
    from azure.storage.blob import BlobPrefix
except (ImportError, AttributeError):
    BlobPrefix = AZURE_BLOB_STORAGE_NOT_IMPORTED  # type: ignore[misc] # assigning to type

try:
    from azure.storage.blob import BlobServiceClient
except (ImportError, AttributeError):
    BlobServiceClient = AZURE_BLOB_STORAGE_NOT_IMPORTED  # type: ignore[misc] # assigning to type

try:
    from azure.storage.blob import ContainerClient
except (ImportError, AttributeError):
    ContainerClient = AZURE_BLOB_STORAGE_NOT_IMPORTED  # type: ignore[misc] # assigning to type



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Type, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.expectations.expectation import (
    ColumnAggregateExpectation,
    parse_value_to_observed_type,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_DESCRIPTION,
    VALUE_SET_DESCRIPTION,
)
from great_expectations.expectations.model_field_types import (
    ValueSetField,  # noqa: TC001  # type needed in pydantic validation
)
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedStringTemplateContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.renderer.observed_value_renderer import ObservedValueRenderState
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = "Expect the set of distinct column values to contain a given set."
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.UNIQUENESS.value]


class ExpectColumnDistinctValuesToContainSet(ColumnAggregateExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnDistinctValuesToContainSet is a \
    Column Aggregate Expectation.

    Column Aggregate Expectations are one of the most common types of Expectation.
    They are evaluated for a single column, and produce an aggregate Metric, such as a mean, standard deviation, number of unique values, column type, etc.
    If that Metric meets the conditions you set, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        value_set (set-like): \
            {VALUE_SET_DESCRIPTION}

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See Also:
        [ExpectColumnDistinctValuesToBeInSet](https://greatexpectations.io/expectations/expect_column_distinct_values_to_be_in_set)
        [ExpectColumnDistinctValuesToEqualSet](https://greatexpectations.io/expectations/expect_column_distinct_values_to_equal_set)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1       1
            1 	2       1
            2 	4       1

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnDistinctValuesToContainSet(
                    column="test",
                    value_set=[1, 4]
                )

            Output:
                {{
                "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                }},
                "result": {{
                    "observed_value": [
                    1,
                    2,
                    4
                    ],
                    "details": {{
                    "value_counts": [
                        {{
                        "value": 1,
                        "count": 1
                        }},
                        {{
                        "value": 2,
                        "count": 1
                        }},
                        {{
                        "value": 4,
                        "count": 1
                        }}
                    ]
                    }}
                }},
                "meta": {{}},
                "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnDistinctValuesToContainSet(
                    column="test2",
                    value_set=[3, 2, 4]
                )

            Output:
                {{
                "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                }},
                "result": {{
                    "observed_value": [
                    1
                    ],
                    "details": {{
                    "value_counts": [
                        {{
                        "value": 1,
                        "count": 3
                        }}
                    ]
                    }}
                }},
                "meta": {{}},
                "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    value_set: ValueSetField

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column aggregate expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    _library_metadata = library_metadata

    # Setting necessary computation metric dependencies and defining kwargs, as well as assigning kwargs default values\  # noqa: E501 # FIXME CoP
    metric_dependencies = ("column.value_counts",)
    success_keys = ("value_set",)

    args_keys = (
        "column",
        "value_set",
    )

    class Config:
        title = "Expect column distinct values to contain set"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnDistinctValuesToContainSet]
        ) -> None:
            ColumnAggregateExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @override
    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("value_set", RendererValueType.ARRAY),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params
        template_str = ""

        if params.value_set:
            array_param_name = "value_set"
            param_prefix = "v__"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            value_set_str: str = cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            template_str = f"distinct values must contain this set: {value_set_str}."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @override
    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> List[RenderedStringTemplateContent]:
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        if renderer_configuration.configuration is None:
            raise ValueError("renderer_configuration.configuration is None.")  # noqa: TRY003 # FIXME CoP
        params = substitute_none_for_missing(
            renderer_configuration.configuration.kwargs,
            [
                "column",
                "value_set",
                "row_condition",
                "condition_parser",
            ],
        )

        if params["value_set"] is None or len(params["value_set"]) == 0:
            values_string = "[ ]"
        else:
            for i, v in enumerate(params["value_set"]):
                params[f"v__{i!s}"] = v

            values_string = " ".join([f"$v__{i!s}" for i, v in enumerate(params["value_set"])])

        template_str = f"distinct values must contain this set: {values_string}."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        styling = runtime_configuration.get("styling", {}) if runtime_configuration else {}

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    @override
    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        observed_value_counts = metrics["column.value_counts"]
        observed_value_set = set(observed_value_counts.index)
        value_set = self._get_success_kwargs()["value_set"]

        # Try to coerce string values to match the type of observed values
        if observed_value_set and value_set:
            first_observed = next(iter(observed_value_set))
            expected_value_set = {
                parse_value_to_observed_type(first_observed, value) for value in value_set
            }
        else:
            expected_value_set = set(value_set)

        if not expected_value_set:
            success = True
        else:
            success = expected_value_set.issubset(observed_value_set)

        return {
            "success": success,
            "result": {
                "observed_value": sorted(list(observed_value_set)),
                "details": {"value_counts": observed_value_counts},
            },
        }

    @classmethod
    @renderer(renderer_type=AtomicDiagnosticRendererType.OBSERVED_VALUE)
    @override
    def _atomic_diagnostic_observed_value(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> RenderedAtomicContent:
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        expected_param_prefix = "exp__"
        expected_param_name = "expected_value"
        ov_param_prefix = "ov__"
        ov_param_name = "observed_value"

        renderer_configuration.add_param(
            name=expected_param_name,
            param_type=RendererValueType.ARRAY,
            value=renderer_configuration.kwargs.get("value_set", []),
        )
        renderer_configuration = cls._add_array_params(
            array_param_name=expected_param_name,
            param_prefix=expected_param_prefix,
            renderer_configuration=renderer_configuration,
        )

        renderer_configuration.add_param(
            name=ov_param_name,
            param_type=RendererValueType.ARRAY,
            value=result.get("result", {}).get("observed_value", []) if result else [],
        )
        renderer_configuration = cls._add_array_params(
            array_param_name=ov_param_name,
            param_prefix=ov_param_prefix,
            renderer_configuration=renderer_configuration,
        )

        observed_value_set = set(
            result.get("result", {}).get("observed_value", []) if result else []
        )

        observed_values = (
            (name, sch)
            for name, sch in renderer_configuration.params
            if name.startswith(ov_param_prefix)
        )
        expected_values = (
            (name, sch)
            for name, sch in renderer_configuration.params
            if name.startswith(expected_param_prefix)
        )

        template_str_list = []
        for name, schema in observed_values:
            renderer_configuration.params.__dict__[
                name
            ].render_state = ObservedValueRenderState.EXPECTED.value
            template_str_list.append(f"${name}")

        for name, schema in expected_values:
            # try to coerce the expected value to a type that can be compared with observed values
            if observed_value_set:
                sample_observed_value = next(iter(observed_value_set))
                expected_value = parse_value_to_observed_type(
                    observed_value=sample_observed_value, value=schema.value
                )
            else:
                expected_value = schema.value

            if expected_value not in observed_value_set:
                renderer_configuration.params.__dict__[
                    name
                ].render_state = ObservedValueRenderState.MISSING.value
                template_str_list.append(f"${name}")

        renderer_configuration.template_str = " ".join(template_str_list)

        value_obj = renderedAtomicValueSchema.load(
            {
                "template": renderer_configuration.template_str,
                "params": renderer_configuration.params.dict(),
                "meta_notes": renderer_configuration.meta_notes,
                "schema": {"type": "com.superconductive.rendered.string"},
            }
        )
        return RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value=value_obj,
            value_type="StringValueType",
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import great_expectations.expectations as gxe
from great_expectations.render.renderer.content_block.content_block import (
    ContentBlockRenderer,
)

if TYPE_CHECKING:
    from great_expectations.expectations.expectation import Expectation


@pytest.mark.parametrize(
    "expectation, expected",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeBetween(
                column="foo", min_value=1, max_value=10, notes="these are _notes_"
            ),
            [
                {
                    "content_block_type": "markdown",
                    "markdown": "these are _notes_",
                    "styling": {"parent": {"styles": {"color": "red"}}},
                },
            ],
            id="string notes",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeBetween(
                column="foo",
                min_value=1,
                max_value=10,
                notes=["I", "have", "**notes**"],
            ),
            [
                {
                    "content_block_type": "markdown",
                    "markdown": "I",
                    "styling": {"parent": {}},
                },
                {
                    "content_block_type": "markdown",
                    "markdown": "have",
                    "styling": {"parent": {}},
                },
                {
                    "content_block_type": "markdown",
                    "markdown": "**notes**",
                    "styling": {"parent": {}},
                },
            ],
            id="list notes",
        ),
    ],
)
@pytest.mark.unit
def test__render_expectation_notes_with_notes(expectation: Expectation, expected: list[dict]):
    config = expectation.configuration
    content = ContentBlockRenderer._render_expectation_notes(config)
    notes_block = content.collapse[0]
    actual = notes_block.to_json_dict()["text"]
    assert actual == expected


@pytest.mark.unit
def test__render_expectation_notes_without_notes():
    expectation = gxe.ExpectColumnValuesToBeBetween(column="foo", min_value=1, max_value=10)
    config = expectation.configuration
    assert ContentBlockRenderer._render_expectation_notes(config) is None
